# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import StOptGrids 
import StOptReg
import StOptGlobal
import Utils
import Simulators as sim
import Optimizers as opt
import dp.DynamicProgrammingByRegressionDist as dyn
import dp.SimulateRegressionDist as srt
import importlib

accuracyClose = 0.2
accuracyEqual = 1e-10


class testSwingOptimSimuMpiTest(unittest.TestCase) :
    
    def test_resolutionSwingOptimSimuMpi(self):
    
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            initialValues = np.zeros(1) + 1.
            sigma = np.zeros(1) + 0.2
            mu = np.zeros(1) + 0.05
            corr = np.ones((1,1))
            # number of step
            nStep = 20
            # exercise date
            dates = np.linspace(0., 1., nStep + 1)
            N = 3 # number of exercises
            T = 1.
            strike = 1.
            nbSimul = 100000
            nMesh = 8
            # payoff for swing
            payOffBasket = Utils.BasketCall(strike);
            payoff = Utils.PayOffSwing(payOffBasket,N)
            nbMesh = np.zeros(1, dtype = np.int32) + nMesh
            # Grid
            lowValues = np.zeros(1, dtype = float)
            step = np.zeros(1, dtype = float) + 1.
            nbStep = np.zeros(1, dtype = np.int32) + N
            grid =  StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
            # initial values
            initialStock = np.zeros(1)
            initialRegime = 0
            fileToDump = "CondExpSwingOptSimHLMpi"
            optimizer = opt.OptimizerSwingBlackScholes(payOffBasket, N)
            bOneFile = True
            # simulator
            simulatorBack = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
            # regressor
            regressor = StOptReg.LocalLinearRegression(nbMesh)
            # link the simulations to the optimizer
            optimizer.setSimulator(simulatorBack)
            # bermudean value
            valueOptimMpi = dyn.DynamicProgrammingByRegressionDist(grid, optimizer, regressor, payoff, initialStock, initialRegime, fileToDump, bOneFile)
            world.barrier()
            # simulation value
            simulatorForward = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, True)
            optimizer.setSimulator(simulatorForward)
            valSimuOneFile = srt.SimulateRegressionDist(grid, optimizer, payoff, initialStock, initialRegime, fileToDump, bOneFile)
            print("Optim", valueOptimMpi, "SIMU", valSimuOneFile)
            
            if world.rank == 0:
                self.assertAlmostEqual(valueOptimMpi, valSimuOneFile, None, None, accuracyClose)
            
            optimizer = opt.OptimizerSwingBlackScholes(payOffBasket, N)
            bOneFile = False
            # simulator
            simulatorBack = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
            # regressor
            regressor = StOptReg.LocalLinearRegression(nbMesh)
            # link the simulations to the optimizer
            optimizer.setSimulator(simulatorBack)
            # bermudean value
            valueOptimMpi = dyn.DynamicProgrammingByRegressionDist(grid, optimizer, regressor, payoff, initialStock, initialRegime, fileToDump, bOneFile)
             
            if world.rank == 0:
                print("Value in optimization", valueOptimMpi)
             
            world.barrier()
            # simulation value
            simulatorForward = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, True)
            optimizer.setSimulator(simulatorForward)
            valSimuMultipleFiles = srt.SimulateRegressionDist(grid, optimizer, payoff, initialStock, initialRegime, fileToDump, bOneFile)
            print("Optim", valueOptimMpi, "SIMU", valSimuMultipleFiles)
            self.assertAlmostEqual(valSimuOneFile, valSimuMultipleFiles, None, None, accuracyEqual)
            
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import matplotlib.pyplot as plt
import time

import StOptGrids
import StOptReg
import StOptGeners

import microgridDEA.simulateState_new as simX
import microgridDEA.valueNext_new as fv
import microgridDEA.sobol_lib as sl


# returns the value function at the final step. Depending upon the type of RMC method, the size of the matrix may be different.
def finalStepValue(param,regParams,demand,inventory):

	return fv.finalValue(param,regParams,demand,inventory)

# creates the matrix using residual demand and inventory for the StOpt function.
def createXMatrix(param,regParams,asset, inventory=[]):

	if regParams.rmctype == 'regress now 2D':
		xMatrix = np.zeros((2,len(asset)))
		xMatrix[0,:] = asset
		xMatrix[1,:] = inventory
		return xMatrix

	if regParams.rmctype == 'gd':
		xMatrix = np.zeros((1,len(asset)))
		xMatrix[0,:] = asset
		return xMatrix

def gridType(param,regParams):

	if regParams.rmctype == 'regress now 2D':
		# notice the grid is only for the regime
		lowValues =np.array([0.],dtype=np.float64)				            # low value for the mesh
		step = np.array([1],dtype=np.float64)  								# size of the mesh
		nbStep = np.array([param.H], dtype=np.int32)						# number of step
		grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)	

	if regParams.rmctype == 'gd':
		# notice the grid is for both the regime and the inventory levels.
		stepI = param.I_minMax[1]/regParams.meshI										
		lowValues =np.array([param.I_minMax[0] , 0.],dtype=np.float64)				
		step = np.array([stepI , 1],dtype=np.float64)  								
		nbStep = np.array([regParams.meshI , param.H], dtype=np.int32)					
		grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)		

	return grid

def regObject(param,regParams,x):

	# rmc type and regression type. The choice is made in the "basicVariables" and "regressionType" class. 
	if regParams.rmctype == 'gd':
		nbMesh = np.array([regParams.meshX],dtype=np.int32)
		regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)

	elif regParams.rmctype == 'regress now 2D':
		if regParams.regType == 'globalPolynomial':
			regressor = StOptReg.GlobalCanonicalRegression(False,x,regParams.degree)

	return regressor


def designSites(param,regParams):

	if regParams.rmctype == 'regress now 2D':
		dim = 2
		# number of unique design sites.
		numdesignSites = regParams.nbUniqueSite 
		skip=0 
		# use sobol sequence for the design sites. This could be replaced by LHS design. 
		sites = sl.i4_sobol_generate(dim, numdesignSites, skip)

		# sobol sequences are generated on square domain. extend it to the domain of interest.
		X0 = -param.ampl + 2*param.ampl*sites[0,:]
		I0 = param.I_minMax[0] + (param.I_minMax[1]-param.I_minMax[0])*sites[1,:]
		
		# repeat the design sites depending upon the batch size.
		X0_rep = np.repeat(X0,regParams.batchSize)
		I0_rep = np.repeat(I0,regParams.batchSize)

		return X0,I0, X0_rep, I0_rep

	elif regParams.rmctype == 'gd':
		dim = 1 # only one dimension because of rmc type.

		numdesignSites = regParams.nbUniqueSite # number of unique design sites
		skip=0
		sites = sl.i4_sobol_generate(dim, numdesignSites, skip ) 

		# sobol sequence in demand dimension and uniformly spaced in the inventory. 		
		X0 = -param.ampl + 2*param.ampl*sites[0,:]
		I0 = np.linspace(param.I_minMax[0],param.I_minMax[1],regParams.meshI+1)

		# repeat the design sites depending upon the batch size.
		X0_rep = np.repeat(X0,regParams.batchSize)

		return X0,I0,X0_rep,I0


def storageCalculation(param, regParamObjcs):

	# parameters of the regression for example: local linear or global polynomial and type of grid.
	regParams = regParamObjcs[-1]

	# create design sites
	X0, inventory, X0_rep, inventory_rep = designSites(param,regParams) 

	# generate one step residual demand
	residualDemand = simX.residualDemand(param, 1, regParams.nbsimulOpt, param.dt, X0_rep, regParams.rmctype)

	# compute the value at the final step. 
	valueNext = finalStepValue(param,regParams,residualDemand[:,1],inventory)

	# average the value for each batch
	r,c = valueNext.shape
	valueNext = np.mean(np.reshape(valueNext, (regParams.batchSize,	regParams.nbUniqueSite, c), order='F'),0)

	archiveToWrite = StOptGeners.BinaryFileArchive(param.filetoDump,"w")


	# iterate on time steps
	for iSteps in range(param.nstep-1,-1,-1):
		print("iSteps",iSteps)

		# create grid for the regression
		grid = gridType(param,regParamObjcs[iSteps])
		
		# create regressor object
		regressor = regObject(param,regParamObjcs[iSteps],createXMatrix(param,regParams,X0,inventory))
		
		# Dump the continuation values in the archive:
		archiveToWrite.dumpGridAndRegressedValue("toStore", param.nstep - iSteps, [valueNext], regressor,grid)

		if iSteps>0:
			regParams = regParamObjcs[iSteps-1]

			# create design sites
			X0, inventory, X0_rep, inventory_rep = designSites(param,regParams) # isites is useless if we use rmctype = 'gd'.

			# generate one step residual demand
			residualDemand = simX.residualDemand(param, 1, regParams.nbsimulOpt, param.dt, X0_rep, regParams.rmctype)
			
			# Read the regressed values
			archiveToRead =  StOptGeners.BinaryFileArchive(param.filetoDump,"r")
			contValues = archiveToRead.readGridAndRegressedValue(param.nstep - iSteps,"toStore")

			# regParamObjcs[iSteps] tells the function how the regression was done and regParamObjcs[iSteps-1] tell the function how to store the valueNext now.
			valueNext, control_d, StOutput = fv.optimization(contValues[0], residualDemand[:,1], inventory_rep, param, regParamObjcs[iSteps-1], regParamObjcs[iSteps])
			
			# average the value for each batch
			r,c = valueNext.shape
			valueNext = np.mean(np.reshape(valueNext, (regParams.batchSize,	regParams.nbUniqueSite, c), order='F'),0)



	



# Copyright (C) 2021 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

import numpy as np
import math
import StOptReg as reg
import StOptGrids
import StOptGlobal
import Utils
import Simulators as sim
import Optimizers as opt
import dp.DynamicProgrammingSwitchingByRegressionHighLevel  as dyn
import dp.SimulateSwitchingRegression  as simOptCont
import unittest
import importlib


# valorisation of a thermal asset
    
class testThermalASsetHighLevelTest(unittest.TestCase):

    def testThermalAsset(self):
        # test MPI
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
        T = 100.
        # number of time step  ( step of 4 hours : 6 step per day)
        nStepPerday = 6
        nbDays = int(T)
        nstep = nStepPerday * nbDays;
        # define a a time grid
        timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., T / nstep, nstep)
  
        #to store the two assets
        futGrids = []
        futValues =  [ 30. + 5 * np.cos((2 * math.pi * i) / nstep) + np.cos(2 * math.pi * i / (7.*nStepPerday)) for i in range(nstep+1)]
        futGrids.append(Utils.FutureCurve(timeGrid, futValues))
        futValues = [ 30. + 5 * np.cos((2 * math.pi * i) / nstep) + np.cos(2 * math.pi* i / (7.*nStepPerday))  + 5.*np.cos(2 * math.pi * i / nStepPerday)   for i in range(nstep+1)]    
        futGrids.append(Utils.FutureCurve(timeGrid, futValues))
        
        sigma = [0.02,0.08] # volatility by product
        mr = [0.004,0.01] # mean reverting by product
        correl = np.identity(2)
        correl[0, 1] = 0.8
        correl[1, 0] = 0.8

        # Constraints
        nbMinOn = 6  #  minimal number of time step the  thermal asset is on
        nbMinOff = 4 #  maximal number of time step the terhma asset is off

        # grids
        gridPerReg =  [ StOptGrids.RegularSpaceIntGrid(np.zeros([1]), (nbMinOn-1)*np.ones([1])),
                        StOptGrids.RegularSpaceIntGrid(np.zeros([1]), (nbMinOff-1)*np.ones([1]))]         

        
        iRegInit = 0 # on at begininng
        pointState = np.zeros([1]) # on since previous time step
        
         # create Thermal Asset object
        switchOffToOnCost = 4
        
        
        # optimizer
        thermalAsset = opt.OptimizeThermalAssetMeanReverting1DAssets(switchOffToOnCost)

        # simulator
        nbSimulOpt =5000
        iSeed = 1
        simulator = sim.MeanReverting1DAssetsSimulator(futGrids, mr, sigma, correl, T , nstep, nbSimulOpt, False,"ArchiveSim", iSeed)

        # regressor
        nMesh = 2
        nbMesh = np.zeros(2, dtype = np.int32) + nMesh
        regressor = reg.LocalLinearRegression(nbMesh)

        # affect simulator to optimizer
        thermalAsset.setSimulator(simulator)
        
        # optimize
        valOptim = dyn.DynamicProgrammingSwitchingByRegressionHighLevel(gridPerReg, thermalAsset, regressor, pointState,iRegInit, "BellmanThermal")
        
        print("valOptimp", valOptim)

        # revert simulator
        iSeed = 10
        nbSimulForward = 5000
        simulator = sim.MeanReverting1DAssetsSimulator(futGrids, mr, sigma, correl, T , nstep, nbSimulForward, True,"ArchiveSim", iSeed)
      
        # affect simulator to optimizer
        thermalAsset.setSimulator(simulator)

        # simulate
        valSimu = simOptCont.SimulateSwitchingRegression(gridPerReg,thermalAsset,pointState,iRegInit, "BellmanThermal")
        print("val simulation", valSimu)

        
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import StOptGrids
import StOptReg
import StOptGlobal
import StOptGeners
import Utils
import Simulators as sim
import Optimizers as opt
import importlib

# unit test for global shape
############################

class OptimizerConstructionMpi(unittest.TestCase):

    def test(self):
        # test MPI
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            initialValues = np.zeros(1,dtype=float) + 1.
            sigma = np.zeros(1) + 0.2
            mu = np.zeros(1) + 0.05
            corr = np.ones((1,1),dtype=float)
             # number of step
            nStep = 30
            # exercice date
            dates = np.linspace(0., 1., nStep + 1)
            T= dates[len(dates) - 1]
            nbSimul = 10 # simulation number (optimization and simulation)
            # simulator
            ##########
            bsSim = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, T, len(dates) - 1, nbSimul, False)
            strike = 1.
            # Pay off
            payOff= Utils.BasketCall(strike)
            # optimizer
            ##########
            N = 3   # number of exercize dates
            swiOpt = opt.OptimizerSwingBlackScholes(payOff,N)
            # link simulator to optimizer
            swiOpt.setSimulator(bsSim)
            # archive
            ########
            nameArchive = "Archive" + str(comm.rank)
            ar = StOptGeners.BinaryFileArchive(nameArchive,"w+")
            # regressor
            ##########
            nMesh = np.array([1])
            regressor = StOptReg.LocalLinearRegression(nMesh)
            # Grid
            ######
            # low value for the meshes
            lowValues =np.array([0.],dtype=float)
            # size of the meshes
            step = np.array([1.],dtype=float)
            # number of steps
            nbStep = np.array([N], dtype=np.int32)
            gridArrival =  StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)
            gridStart   =  StOptGrids.RegularSpaceGrid(lowValues,step,nbStep-1)
            # pay off function for swing
            ############################
            payOffBasket = Utils.BasketCall(strike);
            payoff = Utils.PayOffSwing(payOffBasket,N)
            # final step
            ############
            asset =bsSim.getParticles()
            bSplit = np.array([1],dtype=bool) #define which direction to split in parallel computing
            fin = StOptGlobal.FinalStepDPDist(gridArrival,1,bSplit)
            values = fin.set( payoff,asset)
            # transition time step
            #####################
            # on step backward and get  asset
            asset = bsSim.stepBackwardAndGetParticles()
            # update regressor
            regressor.updateSimulations(0,asset)
            transStep = StOptGlobal.TransitionStepRegressionDPDist(gridStart,gridArrival,swiOpt)
            valuesNext=transStep.oneStep(values,regressor)
            bOneFile=0 # multiple files
            transStep.dumpContinuationValues(ar,"Continuation",1,valuesNext[0],valuesNext[1],regressor,bOneFile)
            # simulate time step
            ####################
            nbSimul= 10
            vecOfStates =[] # state of each simulation
            for i in np.arange(nbSimul):
                # one regime, all with same stock level (dimension 2), same realization of simulation (dimension 3)
                vecOfStates.append(StOptGlobal.StateWithStocks(1, np.array([0.]) , np.zeros(1)))
            simStep = StOptGlobal.SimulateStepRegressionDist(ar,1,"Continuation",gridArrival,swiOpt,bOneFile)
            phi = np.zeros((1,nbSimul))
            NewState =VecOfStateNext = simStep.oneStep(vecOfStates, phi)


if __name__ == '__main__': 
    unittest.main() 



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import StOptReg as reg
import StOptGrids
import StOptGlobal
import Simulators as sim
import Optimizers as opt
import Utils
import dp.DynamicProgrammingByRegressionHighLevel as dyn
import dp.SimulateRegressionControlHighLevel as srt
import unittest
import importlib

accuracyClose = 0.5


def testGasStorage():

    # test MPI
    moduleMpi4Py=importlib.util.find_spec('mpi4py')
    if (moduleMpi4Py is not None):
        from mpi4py import MPI
        # storage
    ###############
    maxLevelStorage = 360000
    injectionRateStorage = 60000
    withdrawalRateStorage = 45000
    injectionCostStorage = 0.35
    withdrawalCostStorage = 0.35
    switchingCostStorage =  4.

    maturity = 1.
    nstep = 10

    # define a a time grid
    timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
    futValues = []
    # periodicity factor
    iPeriod = 52

    for i in range(nstep + 1):
        futValues.append(50. + 20 * math.sin((math.pi * i * iPeriod) / nstep))

    # define the future curve
    futureGrid = Utils.FutureCurve(timeGrid, futValues)
    # one dimensional factors
    nDim = 1
    sigma = np.zeros(nDim) + 0.94
    mr = np.zeros(nDim) + 0.29
    # number of simulations
    nbsimulOpt = 20000
    # grid
    #####
    nGrid = 40
    lowValues = np.zeros(1, dtype = float)
    step = np.zeros(1, dtype = float) + maxLevelStorage / nGrid
    nbStep = np.zeros(1, dtype = np.int32) + nGrid
    grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)

    # no actualization
    rate=0.
    # a backward simulator
    ######################
    bForward = False
    backSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate,maturity, nstep, nbsimulOpt, bForward)
    # optimizer
    ############
    storage = opt.OptimizeGasStorageSwitchingCostMeanReverting(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage, switchingCostStorage)
    # regressor
    ##########
    nMesh = 4
    nbMesh = np.zeros(1, dtype = np.int32) + nMesh
    regressor = reg.LocalLinearRegression(nbMesh)
    # final value
    vFunction = Utils.ZeroPayOff()

    # initial values
    initialStock = np.zeros(1) + maxLevelStorage
    initialRegime = 0 # here do nothing (no injection, no withdrawal)

    # Optimize
    ###########
    fileToDump = "CondExpGasSwiCostHL"
    # link the simulations to the optimizer
    storage.setSimulator(backSimulator)
    valueOptim = dyn.DynamicProgrammingByRegressionHighLevel(grid, storage, regressor, vFunction, initialStock, initialRegime, fileToDump)
    print("valOP", valueOptim)
    nbsimulSim = 40000
    bForward = True
    forSimulator2 = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate,maturity, nstep, nbsimulSim, bForward)
    storage.setSimulator(forSimulator2)
    valSimu2 = srt.SimulateRegressionControl(grid, storage, vFunction, initialStock, initialRegime, fileToDump)
    print("valSimu2", valSimu2, "valOptim", valueOptim)

class testGasStorageSwitchCostTest(unittest.TestCase):

        def test_gasStorageSwitchCostHighLevel(self):

            testGasStorage()

if __name__ == '__main__':

        unittest.main()














# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np


# call payoff for basket
class BasketCall :
    
    m_strike = 0.

    # Contructor
    # p_strike   strike
    def __init__(self, p_strike) :
        
        self.m_strike = p_strike
    
    # for one simulation, give the basket value
    # p_assets   For the current simulation values of all assets
    def apply(self, p_assets) :
        
        return np.maximum(np.mean(p_assets) - self.m_strike, np.zeros(len(p_assets.transpose())))

    # for all simulations, give the basket values
    # p_assets    values of all assets (one column is a simulation)
    def applyVec(self, p_assets) :
        
        ret = np.zeros(p_assets.shape[1])
        
        return np.maximum(np.mean(p_assets, axis=0) - self.m_strike, np.zeros(len(p_assets.transpose())))
        
# Put payoff for basket
class BasketPut :
    
    m_strike = 0.

    # Constructor
    # p_strike   strike
    def __init__(self, p_strike) :
        
        self.m_strike = p_strike

    # for all simulations, give the basket values
    # p_assets    values of all assets (one column is a simulation)
    def operator(self, p_assets) :
        
        ret = np.zeros(p_assets.shape[1])
        
        return np.maximum(-np.mean(p_assets, axis=0) + self.m_strike, np.zeros(len(p_assets.transpose())))



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np


# Defines a fictitious swing on Ndim stocks : the valorisation will be equal to the valorisation of ndim swing
# where ndim is the number of stocks
class OptimizeFictitiousSwing:
    
    # Constructor
    # p_payoff pay off used
    # p_nPointStock number of stock points
    # p_ndim  dimension of the problem (stock number)
    # p_actu   actualization factor
    def __init__(self, p_payoff, p_nPointStock, p_ndim, p_actu):
        
        self.m_payoff = p_payoff
        self.m_nPointStock = p_nPointStock
        self.m_ndim = p_ndim
        self.m_actu = p_actu
        self.m_actuStep = 1.
        
    # define the diffusion cone for parallelism
    # p_regionByProcessor         region (min max) treated by the processor for the different regimes treated
    # return returns in each dimension the min max values in the stock that can be reached from the grid p_gridByProcessor for each regime
    def getCone(self, p_regionByProcessor):
        
        extrGrid = np.zeros(p_regionByProcessor)
        
        # only a single  exercise
        for id in range(self.m_ndim):
            extrGrid[id][1] += 1.
        
        return extrGrid
    
    # defines a step in optimization
    # Notice that this implementation is not optimal. In fact no interpolation is necessary for this asset.
    # This implementation is for test and example purpose
    # p_grid      grid at arrival step after command
    # p_stock     coordinate of the stock point to treat
    # p_condEsp   conditional expectation operator
    # p_phiIn     for each regime  gives the solution calculated at the previous step ( next time step by Dynamic Programming resolution)
    # return for each regimes (column) gives the solution for each particle (row)
    def stepOptimize(self, p_grid, p_stock, p_condEsp, p_phiIn):
        
        nbSimul = p_condEsp[0].getNbSimul()
        solutionAndControl = list(range(2))
        solutionAndControl[0] = np.zeros((nbSimul, 1))
        solutionAndControl[1] = np.zeros((nbSimul, 1))
        payOffVal = self.m_payoff.applyVec(p_condEsp[0].getParticles())
        bConstraint = False
        # calculation detailed for clarity
        # create interpolator at current stock point
        interpolatorCurrentStock = p_grid.createInterpolator(p_stock)
        # cash flow at current stock and previous step
        cashSameStock = interpolatorCurrentStock.applyVec(p_phiIn[0])
        # conditional expectation at current stock point
        condExpSameStock = self.m_actu * p_condEsp[0].getAllSimulations(interpolatorCurrentStock)
        # number of possibilities for arbitrage with m_ndim stocks
        nbArb = (0x01 << self.m_ndim)
        # stock to add
        stockToAdd = np.zeros(self.m_ndim)
        # store optimal conditional  expectation
        condExpOpt = np.zeros(nbSimul) - 1e6
        
        for j in range(nbArb):
            nbArb += 1
            ires = j
            
            for id in range(1, self.m_ndim + 1)[::-1]:
                id -= 1
                idec = (ires >> id)
                stockToAdd[id] = idec
                ires -= (idec << id)
                
            p_stock = p_stock.reshape(2)
            # calculation detailed for clarity
            # create interpolator at next stock point accessible
            nextStock = p_stock + stockToAdd
             
            # test that stock is possible
            if np.amax(nextStock) <= self.m_nPointStock:
                interpolatorNextStock = p_grid.createInterpolator(nextStock)
                # cash flow at next stock previous step
                cashNextStock = interpolatorNextStock.applyVec(p_phiIn[0]).squeeze()
                # conditional expectation at next stock
                condExpNextStock = self.m_actu * p_condEsp[0].getAllSimulations(interpolatorNextStock).squeeze()
                # quantity exercised
                qExercized = stockToAdd.sum()
                # arbitrage
                condExp = payOffVal * qExercized + condExpNextStock
                solutionAndControl[0][:,0] = np.where(condExp > condExpOpt, payOffVal * qExercized + self.m_actu * cashNextStock, solutionAndControl[0][:,0])
                condExpOpt = np.where(condExp > condExpOpt, condExp, condExpOpt)
                solutionAndControl[1][:,0] = np.where(condExp > condExpOpt, qExercized, solutionAndControl[1][:,0])
        
        return solutionAndControl
    
    # defines a step in simulation
    # Notice that this implementation is not optimal. In fact no interpolation is necessary for this asset.
    # This implementation is for test and example purpose
    # p_grid          grid at arrival step after command
    # p_continuation  defines the continuation operator for each regime
    # p_state         defines the state value (modified)
    # p_phiInOut      defines the value function (modified)
    def stepSimulate(self, p_grid, p_continuation, p_state, p_phiInOut):
        
        # number of possibilities for arbitrage with m_ndim stocks
        nbArb = (0x01 << self.m_ndim)
        # stock to add
        stockToAdd = np.zeros(self.m_ndim)
        # initialize
        phiAdd = -1e6
        # max of conditional expectation
        expectationMax = -1e6
        ptStockMax  = np.zeros(len(p_state.getPtStock()))
        bStock = (p_state.getPtStock()[0] == p_state.getPtStock()[1])
        payOffVal = self.m_payoff.apply(p_state.getStochasticRealization())
        
        for j in range(nbArb):
            ires = j
            
            for id in range(self.m_ndim - 1)[::-1]:
                idec = (ires >> id)
                stockToAdd[id] = idec
                ires -= (idec << id)
                
            nextStock = p_state.getPtStock() + stockToAdd
            
            # test that stock is possible
            if nextStock.maxCoeff() <= self.m_nPointStock:
                # quantity exercised
                qExercized = stockToAdd.sum()
                continuationValueNext = self.m_actu * p_continuation[0].getValue(nextStock, p_state.getStochasticRealization())
                expectation = payOffVal * qExercized + continuationValueNext
                
                if (expectation > expectationMax):
                    ptStockMax = nextStock
                    phiAdd = payOffVal * qExercized * self.m_actuStep
                    expectationMax = expectation
                    
        p_state.setPtStock(ptStockMax)
        p_phiInOut += phiAdd
        
    # get number of regimes
    def getNbRegime(self):
        
        return 1
    
    def getNbControl(self):
        
        return 1
    
    # get actualization factor
    def getActuStep(self):
        
        return self.m_actuStep
    
    # store the simulator
    def setSimulator(self, p_simulator):
        
        self.m_simulator = p_simulator
        
    # get the simulator back
    def getSimulator(self):
        
        return self.m_simulator
    
    # get size of the  function to follow in simulation
    def getSimuFuncSize(self):
        
        return 1



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
# Some data being constant per mesh in one dimensional setting
class OneDimData:
    
    # Constructor
    # p_grid  One dimensional grid
    # p_values values on the grid
    def __init__(self, p_grid, p_values):
        
        self.m_grid = p_grid
        self.m_values = p_values
        
        if (len(p_values) == p_grid.getNbStep() + 1) == False:
            pass
        
    # get the value interpolated constant per mesh at a given point
    # p_coord   the abscissa
    # return interpolated value
    def get(self, p_coord):
        
        return self.m_values[self.m_grid.getMesh(p_coord)]



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptReg as reg

class TransitionStepRegressionDP:
    
    def __init__(self, p_pGridCurrent, p_pGridPrevious, p_pOptimize):
        
        self.m_pGridCurrent = p_pGridCurrent
        self.m_pGridPrevious = p_pGridPrevious
        self.m_pOptimize = p_pOptimize
        
    def oneStep(self, p_phiIn, p_condExp):
        
        nbRegimes = self.m_pOptimize.getNbRegime()
        phiOut = list(range(nbRegimes))
        nbControl = self.m_pOptimize.getNbControl()
        controlOut = list(range(nbControl))
        
        # only if the processor is working
        if self.m_pGridCurrent.getNbPoints() > 0:
            
            # allocate for solution
            for iReg in range(nbRegimes):
                phiOut[iReg] = np.zeros((p_condExp.getNbSimul(), self.m_pGridCurrent.getNbPoints()))
                
            for iCont in range(nbControl):
                controlOut[iCont] = np.zeros((p_condExp.getNbSimul(), self.m_pGridCurrent.getNbPoints()))
                
            # number of threads
            nbThreads = 1
            
            contVal = []
            
            for iReg in range(len(p_phiIn)):
                contVal.append(reg.ContinuationValue(self.m_pGridPrevious, p_condExp, p_phiIn[iReg]))
            
            # create iterator on current grid treated for processor
            iterGridPoint = self.m_pGridCurrent.getGridIteratorInc(0)
            
            # iterates on points of the grid
            for iIter in range(self.m_pGridCurrent.getNbPoints()):
                
                if iterGridPoint.isValid():
                    pointCoord = iterGridPoint.getCoordinate()
                    # optimize the current point and the set of regimes
                    solutionAndControl = self.m_pOptimize.stepOptimize(self.m_pGridPrevious, pointCoord, contVal, p_phiIn)
                    
                    # copy solution
                    for iReg in range(self.m_pOptimize.getNbRegime()):
                        phiOut[iReg][:,iterGridPoint.getCount()] = solutionAndControl[0][:,iReg]
                    
                    for iCont in range(nbControl):
                        controlOut[iCont][:,iterGridPoint.getCount()] = solutionAndControl[1][:,iCont]
                
                    iterGridPoint.nextInc(nbThreads)
        
        res = []
        res.append(phiOut)
        res.append(controlOut)
        return res



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import StOptReg as reg
import StOptGrids
import StOptGlobal
import Simulators as sim
import Optimizers as opt
import Utils
import dp.DynamicProgrammingByRegressionDist as dynmpi
import dp.SimulateRegressionControlDist as srtmpi
import unittest
import importlib

accuracyClose = 1e5

    
class testGasStorageSwitchCostMpiHighLevelTest(unittest.TestCase):
    
    def testGasStorageSwitchCostMpiHighLevel(self):
        
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI

            world = MPI.COMM_WORLD
            # storage
            ###############
            maxLevelStorage = 360000.
            injectionRateStorage = 60000.
            withdrawalRateStorage = 45000.
            injectionCostStorage = 0.35
            withdrawalCostStorage = 0.35
            switchingCostStorage =  4.

            maturity = 1.
            nstep = 10

            # define a a time grid
            timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
            futValues = []
            # periodicity factor
            iPeriod = 52

            for i in range(nstep + 1):
                futValues.append(50. + 20 * math.sin((math.pi * i * iPeriod) / nstep))

            # define the future curve
            futureGrid = Utils.FutureCurve(timeGrid, futValues)
            # one dimensional factors
            nDim = 1
            sigma = np.zeros(nDim) + 0.94
            mr = np.zeros(nDim) + 0.29
            # number of simulations
            nbsimulOpt = 20000
            # grid
            #####
            nGrid = 40
            lowValues = np.zeros(1, dtype = float)
            step = np.zeros(1, dtype = float) + maxLevelStorage / nGrid
            nbStep = np.zeros(1, dtype = np.int32) + nGrid
            grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)

            # no actualization
            rate=0. 
            # a backward simulator
            ######################
            bForward = False
            backSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate, maturity, nstep, nbsimulOpt, bForward)
            # optimizer
            ############
            storage = opt.OptimizeGasStorageSwitchingCostMeanReverting(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage, switchingCostStorage)
            # regressor
            ##########
            nMesh = 4
            nbMesh = np.zeros(1, dtype = np.int32) + nMesh
            regressor = reg.LocalLinearRegression(nbMesh)
            # final value
            vFunction = Utils.ZeroPayOff()

            # initial values
            initialStock = np.zeros(1) + maxLevelStorage
            initialRegime = 0 # here do nothing (no injection, no withdrawal)

            # Optimize
            ###########
            fileToDump = "CondExpGasSwiCostHLMpi"
            bOneFile = True
            # link the simulations to the optimizer
            storage.setSimulator(backSimulator)
            valueOptimMpi = dynmpi.DynamicProgrammingByRegressionDist(grid, storage, regressor, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
            print("valOP", valueOptimMpi)

            world.barrier()

            nbsimulSim = 40000
            bForward = True
            forSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate,maturity, nstep, nbsimulSim, bForward)
            storage.setSimulator(forSimulator)
            valSimuMpi = srtmpi.SimulateRegressionControlDist(grid, storage, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
            print("valSimuMpi", valSimuMpi)

            if world.rank == 0:
                self.assertAlmostEqual(valueOptimMpi, valSimuMpi, None, "Re-adjust tolerance edge please", accuracyClose)
                print("Optim", valueOptimMpi, "valSimuMpi", valSimuMpi)

            return valueOptimMpi

    def test_switchingVaryingRegimeStorageMpi(self):
        
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            # storage
            ###############
            maxLevelStorage = 360000.
            injectionRateStorage = 60000.
            withdrawalRateStorage = 45000.
            injectionCostStorage = 0.35
            withdrawalCostStorage = 0.35
            switchingCostStorage =  4.
            
            maturity = 1.
            nstep = 10
        
            # define a a time grid
            timeGrid = StOptGrids.OneDimRegularSpaceGrid(0., maturity / nstep, nstep)
            futValues = []
            # periodicity factor
            iPeriod = 52
            
            for i in range(nstep + 1):
                futValues.append(50. + 20 * math.sin((math.pi * i * iPeriod) / nstep))
                
            # define the future curve
            futureGrid = Utils.FutureCurve(timeGrid, futValues)
            # regime values allowed
            #######################
            tvalues = np.zeros(6)
            tvalues[0] = 0.
            tvalues[1] = 1e-3
            tvalues[2] = 1. / 4 + 1e-3
            tvalues[3] = 1. / 2 + 1e-3
            tvalues[4] = 3. / 4. + 1e-3
            tvalues[5] = 1.
            timeRegimes = StOptGrids.OneDimSpaceGrid(tvalues)
            regValues = []
            regValues.append(3)
            regValues.append(3)
            regValues.append(1)
            regValues.append(3)
            regValues.append(2)
            regValues.append(3)
            regime = Utils.RegimeCurve(timeRegimes, regValues)
            # one dimensional factors
            nDim = 1
            sigma = np.zeros(nDim) + 0.94
            mr = np.zeros(nDim) + 0.29
            # number of simulations
            nbsimulOpt = 20000
            # grid
            #####
            nGrid = 40
            lowValues = np.zeros(1, dtype = float)
            step = np.zeros(1, dtype = float) + maxLevelStorage / nGrid
            nbStep = np.zeros(1, dtype = np.int32) + nGrid
            grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
            
            # no actualization
            rate = 0.
            # a backward simulator
            ######################
            bForward = False
            backSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate, maturity, nstep, nbsimulOpt, bForward)
            # optimizer
            ############
            storage = opt.OptimizeGasStorageSwitchingCostMeanReverting(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage, switchingCostStorage, regime)
            # regressor
            ##########
            nMesh = 4
            nbMesh = np.zeros(1, dtype = np.int32) + nMesh
            regressor = reg.LocalLinearRegression(nbMesh)
            # final value
            vFunction = Utils.ZeroPayOff()
            
            # initial values
            initialStock = np.zeros(1) + maxLevelStorage
            initialRegime = 0 # here do nothing (no injection, no withdrawal)
            
            # Optimize
            ###########
            fileToDump = "CondExpGas"
            bOneFile = True
            # link the simulations to the optimizer
            storage.setSimulator(backSimulator)
            valueOptimMpi = dynmpi.DynamicProgrammingByRegressionDist(grid, storage, regressor, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
            print("valOP", valueOptimMpi)
            
            world.barrier()
            
            # a forward simulator
            #####################
            nbsimulSim = 40000
            bForward = True
            forSimulator = sim.MeanRevertingSimulator(futureGrid, sigma, mr, rate, maturity, nstep, nbsimulSim, bForward)
            # link the simulations to the optimizer
            storage.setSimulator(forSimulator)
            valSimuMpi = srtmpi.SimulateRegressionControlDist(grid, storage, vFunction, initialStock, initialRegime, fileToDump, bOneFile)
            print("valSimuMpi", valSimuMpi)
            
            if world.rank == 0:
                self.assertAlmostEqual(valueOptimMpi, valSimuMpi, None, "Re-adjust tolerance edge please", accuracyClose)
                print("valOP", valueOptimMpi, "valSimuMpi", valSimuMpi)

if __name__ == '__main__':
    
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np

# Give an example for swing options  of the optimization object describing a step in optimization and in simulation


# classical swing option with a given number of exercises
class OptimizeSwing:
    
    # Constructor
    # p_payoff pay off used
    # p_nPointStock number of stock points
    # p_actu   actualization factor
    def __init__(self, p_payoff, p_nPointStock, p_actu):
        
        self.m_payoff = p_payoff
        self.m_nPointStock = p_nPointStock
        self.m_actu = p_actu
        self.m_actuStep = 1.
        
    # define the diffusion cone for parallelism
    # p_regionByProcessor             region (min max) treated by the processor for the differents regimes treated
    # returns in each dimension the min max values in the stock that can be reached from the grid p_gridByProcessor for each regime
    def getCone(self, p_regionByProcessor):
        
        extrGrid = np.zeros(p_regionByProcessor)
        #  only a single  exercise
	extrGrid[0][1] += 1
        
        return extrGrid
    
    # permits to actuallize the time (needed fo simulation)
    def incrementActuTimeStep(self):
        
        self.m_actuStep *= self.m_actu
        
    # defines a step in optimization
    # Notice that this implementation is not optimal. In fact no interpolation is necessary for this asset.
    # This implemnetation is for test and example purpose
    # p_grid      grid at arrival step after command
    # p_stock     coordinate of the stock point to treat
    # p_condEsp   continuation values for each regime
    # p_phiIn     for each regime  gives the solution calculated at the previous step ( next time step by Dynamic Programming resolution)
    # return   a pair  :
    #              - for each regimes (column) gives the solution for each particle (row)
    #              - for each control (column) gives the optimal control for each particule (rows)
    def stepOptimize(self, p_grid, p_stock, p_condEsp, p_phiIn, p_simulator):
        
        nbSimul = p_condEsp[0].getNbSimul()
	solutionAndControl = []
        solutionAndControl[0] = np.zeros((nbSimul, 1))
        solutionAndControl[1] = np.zeros((nbSimul, 1))
        payOffVal = self.m_payoff.applyVec(p_condEsp[0].getParticles())
        # create interpolator at current stock point
	interpolatorCurrentStock = p_grid.createInterpolator(p_stock)
        # cash flow at current stock and previous step
	cashSameStock = interpolatorCurrentStock.applyVec(p_phiIn[0])
	# conditional expectation at current stock point
        condExpSameStock = self.m_actu * p_condEsp[0].getAllSimulations(interpolatorCurrentStock)
        
        if p_stock[0] < self.m_nPointStock:
            # calculation detailed for clarity
            # create interapolator at next stock point accessible
	    nextStock = np.zeros(p_stock)
            nextStock[0] += 1
            interpolatorNextStock = p_grid.createInterpolator(nextStock)
            # cash flow at next stock previous step
	    cashNextStock = interpolatorNextStock.applyVec(p_phiIn[0])
            # conditional espectation at next stock
	    condExpNextStock = self.m_actu * p_condEsp[0].getAllSimulations(interpolatorNextStock)
            
	    # arbitrage
	    solutionAndControl[0][:,0] = np.where(payOffVal + condExpNextStock > condExpSameStock, payOffVal + self.m_actu * cashNextStock, self.m_actu * cashSameStock)
            solutionAndControl[1][:,0] = np.where(payOffVal + condExpNextStock > condExpSameStock, 1, 0.)

	else :
            solutionAndControl[0][:,0] = 0.
            solutionAndControl[1][:,0] = 0.

        return solutionAndControl
    
    # defines a step in simulation
    # Notice that this implementation is not optimal. In fact no interpolation is necessary for this asset.
    # This implementation is for test and example purpose
    # p_grid          grid at arrival step after command
    # p_continuation  defines the continuation operator for each regime
    # p_state         defines the state value (modified)
    # p_phiInOut      defines the value function (modified): size number of functions to follow
    def stepSimulate(self, p_grid, p_continuation, p_simulator, p_state, p_phiInOut):
        
        # only if stock not maximal
	if p_state.getPtStock()[0] < self.m_nPointStock:
            payOffVal = self.m_payoff.apply(p_state.getStochasticRealization())
            continuationValue = self.m_actu * p_continuation[0].getValue(p_state.getPtStock(), p_state.getStochasticRealization())
            
            if p_state.getPtStock()[0] < self.m_nPointStock:
                nextStock = np.zeros(p_state.getPtStock())
                nextStock[0] += 1
                continuationValueNext = self.m_actu * p_continuation[0].getValue(nextStock, p_state.getStochasticRealization())
                
                if payOffVal + continuationValueNext > continuationValue:
                    p_state.setPtStock(nextStock)
                    p_phiInOut += payOffVal * self.m_actuStep
                    
    # get number of regimes
    def getNbRegime(self):
        
        return 1

    # number of controls
    def getNbControl(self):
    
        return 1
    
    
    # get actualization factor
    def getActuStep(self):
        
        return self.m_actuStep

    # store the simulator
    def setSimulator(self, p_simulator):
    
        self.m_simulator = p_simulator
    

    # get the simulator back
    def getSimulator(self):
    
        return self.m_simulator
    

    # get size of the  function to follow in simulation
    def getSimuFuncSize(self):
    
        return 1
    




# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)


from __future__ import division
import numpy as np
import math
import matplotlib.pyplot as plt
import time
from datetime import datetime
import microGrid.parameters as bp
import microGrid.calEngine as ce
import microGrid.forwardSimulations as fs


def runSimulation():


	param = bp.basicVariables()

	ce.storageCalculation(param)


if __name__ == "__main__":

	runSimulation()	
	fs.forwardSimulations()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import dp.TransitionStepRegressionDP as trans
import dp.FinalStepDP as final


def DynamicProgrammingByRegression(p_timeChangeGrid, p_grids, p_optimize, p_regressor, p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump) :
    
    # from the optimizer get back the simulation
    simulator = p_optimize.getSimulator()
    currentTime = simulator.getCurrentStep()
    iTime = len(p_timeChangeGrid) - 1
    
    while currentTime < p_timeChangeGrid[iTime] - 1e3 * 2.220446049250313e-16:
        iTime -= 1
        
    gridCurrent = p_grids[iTime]
    # final values
    valuesNext = final.FinalStepDP(gridCurrent, p_optimize.getNbRegime()).operator(p_funcFinalValue, simulator.getParticles())
    gridPrevious = gridCurrent
    
    # iterate on time steps
    for iStep in range(simulator.getNbStep()) :
        asset = simulator.stepBackwardAndGetParticles()
        currentTime = simulator.getCurrentStep()
        
        while currentTime < p_timeChangeGrid[iTime] - 1e3 * 2.220446049250313e-16:
            iTime -= 1
            
        gridCurrent = p_grids[iTime] 
        # conditional expectation operator
        if iStep == (simulator.getNbStep() - 1):
            p_regressor.updateSimulations(True, asset)
        else:
            p_regressor.updateSimulations(False, asset)
        
        # transition object
        transStep = trans.TransitionStepRegressionDP(gridCurrent, gridPrevious, p_optimize)
        valuesAndControl = transStep.oneStep(valuesNext, p_regressor)
        valuesNext = valuesAndControl[0]
        gridPrevious = gridCurrent
        
    # interpolate at the initial stock point and initial regime
    return (gridPrevious.createInterpolator(p_pointStock).applyVec(valuesNext[p_initialRegime])).mean()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math

# Defines a specialization of the RegularSpaceGrid  object in one dimension


# define a grid in one dimension
class OneDimRegularSpaceGrid:
    
#     # Default constructor
#     def __init__(self):
#         
#         return None
     
    # Constructor
    # p_lowValue   minimal value of the grid
    # p_step       step size
    # p_nbStep     number of steps
    def __init__(self, p_lowValue, p_step, p_nbStep):
        
        self.m_lowValue = p_lowValue
        self.m_step = p_step
        self.m_nbStep = p_nbStep
        
    # To a coordinate get back the mesh number
    # p_coord   coordinate
    # return mesh number associated to the coordinate
    def getMesh(self, p_coord):
        
        if (self.m_lowValue <= p_coord) == False:
            pass
        
        return min(int(math.ceil((p_coord - self.m_lowValue) / self.m_step)), self.m_nbStep)
    
    # get back value
    
    def getLowValue(self):
        
        return self.m_lowValue
    
    def getStep(self):
        
        return self.m_step
    
    def getNbStep(self):
        
        return self.m_nbStep



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptReg as reg
import StOptGrids
import StOptGlobal
import dp.DynamicProgrammingByRegression as dyn
import unittest
import simulators.AR1Simulator as ar1
import dp.OptimizeLake as ol
import dp.SimulateRegressionControlUsingControl as simpart



def almost_equal(x, y, eps=1e-5):
    return abs(x - y) <= eps

class ZeroFunction:

    def __init__(self):

        return None

    def set(self, a, b, c):

        return 0.

# valorization of a given Lake on a  grid
# Gain are proportional to what is withdrawed from the storage
# Only inflows are stochastic
# p_grid             the grid
# p_maxLevelStorage  maximum level
# p_mesh             number of mesh
# p_bCheckClose      Do we check if optimization and simulations are close
def lake(p_grid, p_maxLevelStorage, p_mesh, p_bCheckClose):

    # storage
    #########
    withdrawalRateStorage = 1000

    maturity = 1.
    nstep = 10

    # number of simulations
    nbsimulOpt = 8000
    nbsimulSim = 8000

    # inflow model
    D0 = 50 # initial inflow
    m = D0 # average inflow
    sig = 5. # volatility
    mr = 5. # mean reverting
    # a backward simulator
    ######################
    backSimulator = ar1.AR1Simulator(D0, m, sig, mr, maturity, nstep, nbsimulOpt, False)
    # optimizer
    ##########
    storage = ol.OptimizeLake(withdrawalRateStorage)

    # regressor
    ###########
    nbMesh = np.array([], dtype = np.int32)

    if p_mesh > 0:
        nbMesh = np.zeros(1, dtype = np.int32) + p_mesh

    regressor = reg.LocalLinearRegression(nbMesh)
    # final value
    vFunction = ZeroFunction()

    # initial values
    initialStock = np.zeros(1) + p_maxLevelStorage
    initialRegime = 0 # only one regime

    # Optimize
    fileToDump = "CondExpLake"

    # link the simulations to the optimizer
    storage.setSimulator(backSimulator)
    valueOptim = dyn.DynamicProgrammingByRegression(p_grid, storage, regressor , vFunction, initialStock, initialRegime, fileToDump)

    # Forward simulation using the control :
    forSimulator = ar1.AR1Simulator(D0, m, sig, mr, maturity, nstep, nbsimulSim, True)
    storage.setSimulator(forSimulator)
    valueSimulation = simpart.SimulateRegressionControlUsingControl(p_grid, storage, vFunction, initialStock, initialRegime, fileToDump)

    return valueOptim, valueSimulation


class testLakeTest(unittest.TestCase):

    def setUp(self):
        np.random.seed(0)

    # linear interpolation
    def test_simpleStorageLegendreLinear(self):

        # storage
        #########
        maxLevelStorage = 5000
        # grid
        ######
        nGrid = 10
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 1
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
        nbmesh = 4
        accuracyEqual = 100
        optim, simulationUsingControl  = lake(grid, maxLevelStorage, nbmesh, True)

        self.assertAlmostEqual(optim , simulationUsingControl, None, None, accuracyEqual)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math
import utils.OneDimRegularSpaceGrid as rsg
import utils.OneDimData as data
import simulators.MeanRevertingSimulator as mrsim
import dp.OptimizeGasStorage as ogs
import dp.DynamicProgrammingByRegression as dyn
import StOptReg as reg
import StOptGrids
import unittest

accuracyClose = 1.5


class ZeroFunction :
    
    def __init__(self) :
        
        return None

    def set(self, a, b, c) :
        
        return 0.

# valorization of a given gas storage on a  grid
# p_grid             the grid
# p_maxLevelStorage  maximum level
def gasStorage(p_grid, p_maxLevelStorage) :
    
    # storage
    injectionRateStorage = 60000
    withdrawalRateStorage = 45000
    injectionCostStorage = 0.35
    withdrawalCostStorage = 0.35

    maturity = 1.
    nstep = 100

    # define a a time grid
    timeGrid = rsg.OneDimRegularSpaceGrid(np.zeros(1), maturity / nstep, nstep)
    # future values
    futValues = np.zeros(nstep + 1)
    
    # periodicity factor
    iPeriod = 52
    
    for i in list(range(nstep + 1)) :
        futValues[i] = 50. + 20 * math.sin((math.pi * i * iPeriod) / nstep)
        
    # define the future curve
    futureGrid = data.OneDimData(timeGrid, futValues)
    
    # one dimensional factors
    nDim = 1
    sigma = np.zeros(nDim) + 0.94
    mr = np.zeros(nDim) + 0.29
    # number of simulations
    nbsimulOpt = 20000
    
    # no actualization
    rate=0
    # a backward simulator
    bForward = False
    
    backSimulator = mrsim.MeanRevertingSimulator(futureGrid, sigma, mr, rate,maturity, nstep, nbsimulOpt, bForward)
    # optimizer
    storage = ogs.OptimizeGasStorage(injectionRateStorage, withdrawalRateStorage, injectionCostStorage, withdrawalCostStorage)
    
    # regressor
    nMesh = 6
    nbMesh = np.zeros(1, dtype = np.int32) + nMesh
    regressor = reg.LocalLinearRegression(nbMesh)
    # final value
    vFunction = ZeroFunction()
    
    # initial values
    initialStock = np.zeros(1) + p_maxLevelStorage
    initialRegime = 0 # only one regime
    
    # Optimize
    fileToDump = "CondExpGasStorage"
    
    # link the simulations to the optimizer
    storage.setSimulator(backSimulator)
    valueOptim = dyn.DynamicProgrammingByRegression(p_grid, storage, regressor, vFunction, initialStock, initialRegime, fileToDump)
    print("valueOptim", valueOptim)
    return valueOptim
            
class testGasStorageTest(unittest.TestCase):
          
    def test_simpleStorage(self):
              
        # storage
        #########
        maxLevelStorage = 90000
        # grid
        ######
        nGrid = 10
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        grid = StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
              
        gasStorage(grid, maxLevelStorage)
     
    def test_simpleStorageLegendreLinear(self):
         
        # storage
        #########
        maxLevelStorage = 90000
        # grid
        ######
        nGrid = 10
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 1
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
              
        gasStorage(grid, maxLevelStorage)
        
    def test_simpleStorageLegendreQuadratic(self):
        
        # storage
        #########
        maxLevelStorage = 90000
        # grid
        ######
        nGrid = 5
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 2
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
              
        gasStorage(grid, maxLevelStorage)
        
    def test_simpleStorageLegendreCubic(self):
        
        # storage
        #########
        maxLevelStorage = 90000
        # grid
        ######
        nGrid = 5
        lowValues = np.zeros(1)
        step = np.zeros(1) + (maxLevelStorage / nGrid)
        nbStep = np.zeros(1, dtype = np.int32) + nGrid
        poly = np.zeros(1, dtype = np.int32) + 3
        grid = StOptGrids.RegularLegendreGrid(lowValues, step, nbStep, poly)
              
        gasStorage(grid, maxLevelStorage)
        
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random
import math
import StOptGrids
import StOptReg
import StOptGeners
import matplotlib.pyplot as plt

# unit test for dumping binary archive of regressed value and Read then
######################################################################

class testBinaryArchiveStOpt(unittest.TestCase):


    def  createData(self):

        X1=np.arange(-1.0 , 1.0 , 0.01 )
        X2=np.arange(-1.0 , 1.0 , 0.01 )
        Y=np.zeros((len(X1),len(X2)))
        for i in range(len(X1)):
            for j in range(len(X2)):
                if i < len(X1)//2:
                    if j < len(X2)//2:
                        Y[i,j]=X1[i]+X2[j]
                    else:
                       Y[i,j]=4*X1[i]+4*X2[j]
                else:
                    if j < len(X2)//2:
                        Y[i,j]=2*X1[i]+1*X2[j]
                    else:
                        Y[i,j]=2*X1[i]+3*X2[j]

        XX1, XX2 = np.meshgrid(X1,X2)  
        Y=Y.T

        r,c = XX1.shape

        X =  np.reshape(XX1,(r*c,1))[:,0]
        I =  np.reshape(XX2,(r*c,1))[:,0]
        Y =  np.reshape(Y,(r*c,1))[:,0]

        xMatrix = np.zeros((2,len(X)))
        xMatrix[0,:] = X
        xMatrix[1,:] = I

        return xMatrix, Y


    def  createData1(self):

        X1=np.arange(-1.0 , 1.0 , 0.01 )
        X2=np.arange(-1.0 , 1.0 , 0.01 )
        Y=np.zeros((len(X1),len(X2)))
        for i in range(len(X1)):
            for j in range(len(X2)):
                if i < len(X1)//2:
                    if j < len(X2)//2:
                        Y[i,j]=2*X1[i]+5*X2[j]
                    else:
                       Y[i,j]=6*X1[i]+6*X2[j]
                else:
                    if j < len(X2)//2:
                        Y[i,j]=0.5*X1[i]+0.5*X2[j]
                    else:
                        Y[i,j]=X1[i]+10*X2[j]

        XX1, XX2 = np.meshgrid(X1,X2)  
        Y=Y.T

        r,c = XX1.shape

        X =  np.reshape(XX1,(r*c,1))[:,0]
        I =  np.reshape(XX2,(r*c,1))[:,0]
        Y =  np.reshape(Y,(r*c,1))[:,0]

        xMatrix = np.zeros((2,len(X)))
        xMatrix[0,:] = X
        xMatrix[1,:] = I

        return xMatrix, Y


    def testSimpleStorageAndLecture(self):
        
        # i create grid of 5 starting from 0 till 4
        lowValues =np.array([0],dtype=float)
        step = np.array([1],dtype=float)
        nbStep = np.array([4], dtype=np.int32)
        grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)

        xMatrix, y = self.createData()

        xMatrix, y1 = self.createData1()

        # create regressor object
        nbMesh = np.array([2,2],dtype=np.int32)
        regressor = StOptReg.LocalLinearRegression(False,xMatrix,nbMesh)
        
        # create a matrix (number of simulations by number of grid points)
        toRegressMult = np.zeros(shape=(len(y),grid.getNbPoints()))

        for i in range(toRegressMult.shape[1]):
            toRegressMult[:,i] = y

        toRegressMult[:,1] = y1


        # Create the binary archive to dump
        ###################################
        archiveToWrite = StOptGeners.BinaryFileArchive("MyArchive","w")
        # step 1
        archiveToWrite.dumpGridAndRegressedValue("toStore", 1, [toRegressMult], regressor,grid)


        # Read the regressed values
        ###########################
        archiveToRead =  StOptGeners.BinaryFileArchive("MyArchive","r")
        contValues = archiveToRead.readGridAndRegressedValue(1,"toStore")

        # test getMesh1D method
        meshpoints  = regressor.getMesh1D()
                
        
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2018 The Regents of the University of California
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
# Authors - Aditya Maheshwari and Mike Ludkovski
#           Department of Statistics and Applied Probability,
#           University of California at Santa Barbara

from __future__ import division
import numpy as np
import math
import matplotlib
# matplotlib.use('Agg')
import matplotlib.pyplot as plt
import time

import parameters as bv
import simulateState_new as simX
import valueNext_new as fv


import StOptGrids
import StOptReg
import StOptGeners



if __name__ == "__main__":

    outSampleSimul = 10000

    param = bv.basicVariables()

    condExpFilePath = 'condExp_rmcType_gd_regType_piecewiseLinear'
    
    I0 = np.random.uniform(param.I_minMax[0],param.I_minMax[1],(outSampleSimul,1))
    Xt = np.random.uniform(-8,8,(outSampleSimul,1))
    
    r,c = Xt.shape

    It = np.zeros((r,2)) # inventory evolution
    It[:,0] = I0[:,0]

    Dt = np.zeros((r,2), dtype=np.int) # clock 
    Dt[:,0] = 1  #initial regime

    St = np.zeros((r,c)) # stores Xt - dt - Bt
    Bt = np.zeros((r,c)) # battery output
    dt = np.zeros((r,c)) # diesel output

    archiveToRead =  StOptGeners.BinaryFileArchive(condExpFilePath,"r")


    contValues = archiveToRead.readGridAndRegressedValue(20,"toStore")

    for sim in range(outSampleSimul):
        _, dt[sim, 0], St[sim, 0], It[sim, 1] ,Bt[sim, 0] = fv.findOptimalControl(Xt[sim,0], It[sim,0], Dt[sim,0], contValues[0], param)


    plt.scatter(Xt[:,0],It[:,0],c=dt[:,0])
    plt.show()
        



# Copyright (C) 2020 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest

# unit test for fast CDF on grid
################################

class testFastCDF(unittest.TestCase):
    def test2D(self):
        
        import StOptCDF
        nbSimul = 100000
        x = np.random.normal(size=(2,nbSimul));
        # y =1 for simple CDF
        y =  np.ones([nbSimul])

        # define the grid (same in both dimension)
        z1= -2 + 4*np.arange(20)/20
        z = [z1,z1]
        z = StOptCDF.fastCDF(x,z,y)
        
        

if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2021 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptGrids 
import StOptReg
import StOptGlobal
import StOptGeners

def DynamicProgrammingSwitchingByRegressionHighLevel(p_grid, p_optimize, p_regressor, p_pointState, p_initialRegime, p_fileToDump) :
    
    # from the optimizer get back the simulation
    simulator = p_optimize.getSimulator()
    print("simulator", simulator)
    # final values
    valuesNext= []
    for iReg in range(len(p_grid)):
        valuesNext.append( np.zeros([simulator.getNbSimul(),p_grid[iReg].getNbPoints()]))
        
    ar = StOptGeners.BinaryFileArchive(p_fileToDump, "w")
    nameAr = "ContinuationSwitching"
    nsteps =simulator.getNbStep()
    # iterate on time steps
    for iStep in range(nsteps) :
        asset = simulator.stepBackwardAndGetParticles()
        # conditional expectation operator
        if iStep == (simulator.getNbStep() - 1):
            p_regressor.updateSimulations(True, asset)
        else:
            p_regressor.updateSimulations(False, asset)
        
        # transition object
        transStep = StOptGlobal.TransitionStepRegressionSwitch(p_grid, p_grid, p_optimize)
        values = transStep.oneStep(valuesNext, p_regressor)
        transStep.dumpContinuationValues(ar, nameAr, nsteps - 1 -iStep, valuesNext, p_regressor)
        valuesNext = values
        
    # interpolate at the initial stock point and initial regime
    return valuesNext[p_initialRegime][:,p_grid[p_initialRegime].globCoordPerDimToLocal( p_pointState)].mean()



# Copyright (C) 2016, 2018 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptReg
import StOptGeners
import dp.TransitionStepRegressionDP as trans
import dp.FinalStepDP as final


def DynamicProgrammingByRegression(p_grid, p_optimize, p_regressor, p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump, key1="Continuation" , key2 = "Control"):

    # from the optimizer get back the simulation
    simulator = p_optimize.getSimulator()
    # final values
    valuesNext = final.FinalStepDP(p_grid, p_optimize.getNbRegime()).operator(p_funcFinalValue, simulator.getParticles())

    archiveToWrite = StOptGeners.BinaryFileArchive(p_fileToDump, "w")
    nsteps = simulator.getNbStep()
    # iterate on time steps
    for iStep in range(nsteps):
        asset = simulator.stepBackwardAndGetParticles()

        # conditional expectation operator
        if iStep == (simulator.getNbStep() - 1):
            p_regressor.updateSimulations(True, asset)
        else:
            p_regressor.updateSimulations(False, asset)

        # transition object
        transStep = trans.TransitionStepRegressionDP(p_grid, p_grid, p_optimize)
        valuesAndControl = transStep.oneStep(valuesNext, p_regressor)
        # Dump the continuation values in the archive:
        control =  valuesAndControl[1]
        archiveToWrite.dumpGridAndRegressedValue(key1, nsteps - 1 - iStep, valuesNext, p_regressor, p_grid)
        archiveToWrite.dumpGridAndRegressedValue(key2, nsteps - 1 - iStep, control, p_regressor, p_grid)
        valuesNext = valuesAndControl[0]


    # interpolate at the initial stock point and initial regime
    return (p_grid.createInterpolator(p_pointStock).applyVec(valuesNext[p_initialRegime])).mean()



# Copyright (C) 2017 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import math as maths
import utils.BasketOptions as bo
import simulators.BlackScholesSimulator as bs
import utils.americanOption as amer    
import StOptGrids
import StOptReg as reg
import unittest



def americanKernel(p_nDim, p_nbSimul ,regressor):
    
    initialValues = np.zeros(p_nDim) + 1.0
    sigma = np.zeros(p_nDim) + 0.2
    mu = np.zeros(p_nDim) + 0.05
    corr = np.zeros((p_nDim,p_nDim))
    T = 1.
    nDate = 10
    np.fill_diagonal(corr, 1.)
    strike = 1.

    # simulator
    simulator = bs.BlackScholesSimulator(initialValues, sigma, mu, corr, T, nDate, p_nbSimul, False)
    # payoff
    payoff = bo.BasketPut(strike)

    # bermudean value
    value = amer.resolution(simulator, payoff, regressor)
    print("value",value)
    return value

# test cases
class testAmericanKernelOptionTest(unittest.TestCase):
    
    def test_americanBasketKernelLin(self):
        # dimension
        nDim = 1
        nbSimul = 5000
        bandWidth = 0.1
        referenceValue = 0.06031
        accuracyEqual = 1.
        # regressor
        regressor = reg.LocalGridKernelRegression(bandWidth,1.,True)
        self.assertAlmostEqual(americanKernel(nDim, nbSimul, regressor), referenceValue, None, None, accuracyEqual)

    def test_americanBasketKernelConst(self):
        # dimension
        nDim = 1
        nbSimul = 5000
        bandWidth = 0.1
        referenceValue = 0.06031
        accuracyEqual = 1.
        # regressor
        regressor = reg.LocalGridKernelRegression(bandWidth,1.,False)
        self.assertAlmostEqual(americanKernel(nDim, nbSimul, regressor), referenceValue, None, None, accuracyEqual)



if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 EDF
# Copyright (C) 2020 CSIRO Data61
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random

# unit test for Laplacian  kernel regression 
############################################

class testLaplacianKernelRegression(unittest.TestCase):

    # One dimensional test
    def testRegressionConst1D(self):
        import StOptReg 
        nbSimul = 100000;
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # real function
        toReal = (4+x[0,:]+(3+x[0,:])*(2+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # bandwidth
        bandwidth = 0.05*np.ones(1)
        # multiplicative for rectilinear grid
        coeffMul = 2
        # Linear or constant regressions
        bLinear= False
        # Regressor
        regressor = StOptReg.LaplacianGridKernelRegression(False,x,bandwidth, coeffMul,bLinear)
        # nb simul
        nbSimul= regressor.getNbSimul()
        # particles
        part = regressor.getParticles()
        # test particules
        y = regressor.getAllSimulations(toRegress).transpose()
        # compare to real value
        diff = max(abs(y-toReal)/y)
        self.assertAlmostEqual(diff,0., 1,"Difference between function and its condition expectation estimated greater than tolerance")

    def testRegressionLinear1D(self):
        import StOptReg 
        nbSimul = 100000;
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # real function
        toReal = (4+x[0,:]+(3+x[0,:])*(2+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # bandwidth
        bandwidth = 0.05*np.ones(1)
        # multiplicative for rectilinear grid
        coeffMul = 2
        # Linear or constant regressions
        bLinear= True
        # Regressor
        regressor = StOptReg.LaplacianGridKernelRegression(False,x,bandwidth, coeffMul,bLinear)
        # nb simul
        nbSimul= regressor.getNbSimul()
        # particles
        part = regressor.getParticles()
        # test particules
        y = regressor.getAllSimulations(toRegress).transpose()
        # compare to real value
        diff = max(abs(y-toReal)/y)
        self.assertAlmostEqual(diff,0., 1,"Difference between function and its condition expectation estimated greater than tolerance")

    # two dimensonnal test
    def testRegressionConstKernel(self):
        import StOptReg 
        np.random.seed(1000)
        nbSimul = 400000;
        x = np.random.uniform(-1.,1.,size=(2,nbSimul));
        # real function
        toReal = (4+x[0,:])*(3+x[1,:])
        # function to regress
        toRegress = toReal + np.random.normal(0.,1,nbSimul)
        # bandwidth
        bandwidth = 0.05*np.ones(2)
        # multiplicative for rectilinear grid
        coeffMul = 2
        # Linear or constant regressions
        bLinear= False
        # Regressor
        regressor = StOptReg.LaplacianGridKernelRegression(False,x,bandwidth, coeffMul,bLinear)
        # test
        y = regressor.getAllSimulations(toRegress).transpose()
        # particles
        part = regressor.getParticles()
        # compare to real value
        diff = max(abs(y-toReal)/toReal)
        self.assertAlmostEqual(diff,0., 1,"Difference between function and its condition expectation estimated greater than tolerance")

    # two dimensonnal test
    def testRegressionLinearKernel(self):
        import StOptReg 
        np.random.seed(1000)
        nbSimul = 400000;
        x = np.random.uniform(-1.,1.,size=(2,nbSimul));
        # real function
        toReal = (4+x[0,:])*(3+x[1,:])
        # function to regress
        toRegress = toReal + np.random.normal(0.,1,nbSimul)
        # bandwidth
        bandwidth = 0.05*np.ones(2)
        # multiplicative for rectilinear grid
        coeffMul = 2
        # Linear or constant regressions
        bLinear= True
        # Regressor
        regressor = StOptReg.LaplacianGridKernelRegression(False,x,bandwidth, coeffMul,bLinear)
        y = regressor.getAllSimulations(toRegress).transpose()
        # particles
        part = regressor.getParticles()
        # compare to real value
        diff = max(abs(y-toReal)/toReal)
        self.assertAlmostEqual(diff,0., 1,"Difference between function and its condition expectation estimated greater than tolerance")

    # test different second member
    def testSecondMemberConstKernel(self):
         import StOptReg 
         np.random.seed(1000)
         nbSimul = 100;
         x = np.random.uniform(-1.,1.,size=(1,nbSimul));
         toReal = np.array([((2+x[0,:]+(1+x[0,:])*(1+x[0,:]))).tolist(),((3+x[0,:]+(2+x[0,:])*(2+x[0,:]))).tolist()])
         # function to regress
         toRegress = np.array([(toReal[0,:] + 4*np.random.normal(0.,1,nbSimul)).tolist(),(toReal[1,:] + 4*np.random.normal(0.,1,nbSimul)).tolist()])
         # bandwidth
         bandwidth = 0.1*np.ones(1)
         # multiplicative for rectilinear grid
         coeffMul = 2
         # Linear or constant regressions
         bLinear= False
         # Regressor
         regressor = StOptReg.LaplacianGridKernelRegression(False,x,bandwidth, coeffMul,bLinear)
         y = regressor.getAllSimulationsMultiple(toRegress)
         # check 
         z1 = regressor.getAllSimulations(toRegress[0,:]).transpose()
         z2 = regressor.getAllSimulations(toRegress[1,:]).transpose()
         # compare to real value
         diff = max(abs(y[0,:]-z1[:] ))+max(abs(y[1,:]-z2[:] ))
         self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")

    def testSecondMemberLinearKernel(self):
         import StOptReg 
         np.random.seed(1000)
         nbSimul = 100;
         x = np.random.uniform(-1.,1.,size=(1,nbSimul));
         toReal = np.array([((2+x[0,:]+(1+x[0,:])*(1+x[0,:]))).tolist(),((3+x[0,:]+(2+x[0,:])*(2+x[0,:]))).tolist()])
         # function to regress
         toRegress = np.array([(toReal[0,:] + 4*np.random.normal(0.,1,nbSimul)).tolist(),(toReal[1,:] + 4*np.random.normal(0.,1,nbSimul)).tolist()])
         # bandwidth
         bandwidth = 0.1*np.ones(1)
         # multiplicative for rectilinear grid
         coeffMul = 2
         # Linear or constant regressions
         bLinear= True
         # Regressor
         regressor = StOptReg.LaplacianGridKernelRegression(False,x,bandwidth, coeffMul,bLinear)
         y = regressor.getAllSimulationsMultiple(toRegress)
         # check 
         z1 = regressor.getAllSimulations(toRegress[0,:]).transpose()
         z2 = regressor.getAllSimulations(toRegress[1,:]).transpose()
         # compare to real value
         diff = max(abs(y[0,:]-z1[:] ))+max(abs(y[1,:]-z2[:] ))
         self.assertAlmostEqual(diff,0., 7,"Difference between function and its condition expectation estimated greater than tolerance")


    # test  date 0
    def testDateZeroConstKernel(self):
        import StOptReg 
        nbSimul = 50000;
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # real function
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # bandwidth
        bandwidth = 0.05*np.ones(1)
        # multiplicative for rectilinear grid
        coeffMul = 2
        # Linear or constant regressions
        bLinear= False
        # Regressor
        regressor = StOptReg.LaplacianGridKernelRegression(True,x,bandwidth, coeffMul,bLinear)
        y = regressor.getAllSimulations(toRegress).mean()
        self.assertAlmostEqual(y,toRegress.mean(),8,"Not equality by average")

    def testDateZeroLinearKernel(self):
        import StOptReg 
        nbSimul = 50000;
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # real function
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # bandwidth
        bandwidth = 0.05*np.ones(1)
        # multiplicative for rectilinear grid
        coeffMul = 2
        # Linear or constant regressions
        bLinear= True
        # Regressor
        regressor = StOptReg.LaplacianGridKernelRegression(True,x,bandwidth, coeffMul,bLinear)
        y = regressor.getAllSimulations(toRegress).mean()
        self.assertAlmostEqual(y,toRegress.mean(),8,"Not equality by average")
        
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import StOptGrids 
import StOptReg
import StOptGlobal
import Utils
import Simulators as sim
import Optimizers as opt
import dp.DynamicProgrammingByRegressionDist as dynmpi
import dp.SimulateRegressionDist as srtmpi
import importlib

accuracyClose = 1.5


# p_ndim        dimension of the swing
# p_bOneFile    Do we use one unique file for continuation values
def swingND(p_ndim, p_bOneFile):
    
    from mpi4py import MPI
    world = MPI.COMM_WORLD
    initialValues = np.zeros(1) + 1.
    sigma = np.zeros(1) + 0.2
    mu = np.zeros(1) + 0.05
    corr = np.ones((1,1))
    # number of step
    nStep = 20
    # exercise date
    dates = np.linspace(0., 1., nStep + 1)
    N = 3 # number of exercises
    T = 1.
    strike = 1.
    nbSimul = 80000
    nMesh = 8
    # payoff for swing
    payOffBasket = Utils.BasketCall(strike)
    payoff = Utils.PayOffFictitiousSwing(payOffBasket,N)
    nbMesh = np.zeros(1, dtype = np.int32) + nMesh
    # simulator
    simulator = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
    # Grid
    lowValues = np.zeros(p_ndim)
    step = np.zeros(p_ndim) + 1.
    # the stock is discretized with values from 0 to N included
    nbStep = np.zeros(p_ndim, dtype = np.int32) + N
    grid =  StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
    # optimizer
    optimizer = opt.OptimizerFictitiousSwingBlackScholes(payOffBasket, N, p_ndim)
    # initial values
    initialStock = np.zeros(p_ndim)
    initialRegime = 0
    fileToDump = "CondExpSwingOptimNDHLMpi"
    # regressor
    regressor = StOptReg.LocalLinearRegression(nbMesh)
    # link the simulations to the optimizer
    optimizer.setSimulator(simulator)
    # bermudean value
    valueOptimMpi = dynmpi.DynamicProgrammingByRegressionDist(grid, optimizer, regressor, payoff, initialStock, initialRegime, fileToDump, p_bOneFile)
    # simulation value
    simulatorForward = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, True)
    optimizer.setSimulator(simulatorForward)
    valSimuMpi = srtmpi.SimulateRegressionDist(grid, optimizer, payoff, initialStock, initialRegime, fileToDump, p_bOneFile)
    print("Optim", valueOptimMpi, "SIMU", valSimuMpi)
    return valueOptimMpi, valSimuMpi

class testSwingOptimSimuNDMpiTest(unittest.TestCase):
    
    def test_swingOptionOptim2DSimuMpiOneFile(self):
        
       moduleMpi4Py=importlib.util.find_spec('mpi4py')
       if (moduleMpi4Py is not None):
           from mpi4py import MPI
           world = MPI.COMM_WORLD
           ndim = 2
           bOneFile = True
           val = swingND(ndim, bOneFile)
           
           if world.rank == 0:
               self.assertAlmostEqual(val[0], val[1], None, None, accuracyClose)
        
    def test_swingOptionOptim2DSimuMpiMultipleFile(self):
        
      moduleMpi4Py=importlib.util.find_spec('mpi4py')
      if (moduleMpi4Py is not None):
          from mpi4py import MPI
          world = MPI.COMM_WORLD
          ndim = 2
          bOneFile = False
          val = swingND(ndim, bOneFile)
          
          if world.rank == 0:
              self.assertAlmostEqual(val[0], val[1], None, None, accuracyClose)
            
if __name__ == '__main__':
    
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import StOptGrids 
import StOptReg
import StOptGlobal
import Utils
import Simulators as sim
import dp.DynamicProgrammingByRegression as dyn
import dp.FinalValueFictitiousFunction as fvff
import dp.OptimizeFictitiousSwing as OptimizeFictitiousSwing
import utils.BasketOptions as bo

accuracyClose = 1.5


class testSwingOptimSimuNDTest(unittest.TestCase) :
    
    def test_resolutionSwingOptimSimu(self):
    
        ndim = 2
        initialValues = np.zeros(1) + 1.
        sigma = np.zeros(1) + 0.2
        mu = np.zeros(1) + 0.05
        corr = np.ones((1,1))
        # number of step
        nStep = 20
        # exercise date
        dates = np.linspace(0., 1., nStep + 1)
        N = 3 # number of exercises
        T = 1.
        strike = 1.
        nbSimul = 80000
        nMesh = 8
        # payoff for swing
        payOffBasket = bo.BasketCall(strike)
        nbMesh = np.zeros(1, dtype = np.int32) + nMesh
        # simulator
        simulatorBack = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
        # Grid
        lowValues = np.zeros(ndim)
        step = np.zeros(ndim) + 1.
        nbStep = np.zeros(ndim, dtype = np.int32) + N
        grid =  StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
        vFunction = fvff.FinalValueFictitiousFunction(payOffBasket, N)
        # optimizer
        actu = math.exp(-mu[0] * T / nStep)
        optimizer = OptimizeFictitiousSwing.OptimizeFictitiousSwing(payOffBasket, N, ndim, actu)
        # initial values
        initialStock = np.zeros(ndim)
        initialRegime = 0
        fileToDump = "CondExpSwingOptimND"
        # regressor
        regressor = StOptReg.LocalLinearRegression(nbMesh)
        # link the simulations to the optimizer
        optimizer.setSimulator(simulatorBack)
        # bermudean value
        valueOptim = dyn.DynamicProgrammingByRegression(grid, optimizer, regressor, vFunction, initialStock, initialRegime, fileToDump)
        print("Optim", valueOptim)
        
if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np

class basicVariables:


    def __init__(self, maturity=10):

        self.K = 10 #switching cost
        

        self.meshX = 4 #number of meshes on demand dimension
        self.meshI = 10 #number of meshes on inventory dimension

        self.I0=5 #initial inventory

        self.I_minMax = [0,10] #Level of charge of the battery min max in KW
        self.B_minMax =  [-6,6] #battery min max output, positive implies output and negative implies input
        self.d_minMax =  [1,10] #battery min max output, positive implies output and negative implies input


        self.H = 1 # represents the number of regimes which in this case is 2, as a result H=1

        # assuming the cost function is of the type 
        # c1*d^a + c2*1{st>0} + c3*st*1{st<0}
        c =[1,1000000,5] # assuming the cost function is of the type , c[0]*d^a + c[1]*1{st>0} + c[2]*1{st<0}(-st)
        self.a = 0.9
        self.c1 = c[0]
        self.c2 = c[1]
        self.c3 = c[2]


        self.maturity = maturity #unit in hours
        self.dt = 15/60 #unit in hours
        self.nstep = int(self.maturity/self.dt) #number of macro steps

        self.controlType = 'notBangBang' #'BangBang' #

        if self.controlType  != 'BangBang':
            self.steps_d = 10   # discretization of diesel engine output
            self.discrertize_d = np.hstack((0,np.linspace(self.d_minMax[0],self.d_minMax[1],self.steps_d)))[:,None].T

        # demand parameters
        self.lambd = 0.5 #mean reversion rate
        self.sigma = 2 #volatility
        self.ampl = 6 
        self.mu = 0


        self.DemandUB = 10

        self.rmctype = 'regress now 2D' #'gd' #
        
        if self.rmctype == 'gd':
            self.regType = 'piecewiseLinear'
            # number of simulations
            self.nbsimulOpt = 500*self.meshX

        elif self.rmctype == 'regress now 2D':
            self.regType =  'globalPolynomial' #'kernel' #'piecewiseLinear' #
            self.nbsimulOpt = 10000

            self.degree= 2  # used only with regType = 'globalPolynomial'

            # used only with regType = 'kernel'
            self.bandwidth = 0.2
            # factor for the number of points
            self.factPoint=1.

        self.filetoDump = "condExp_rmcType_"+self.rmctype+"_regType_" + self.regType+"_controlType_"+self.controlType



if __name__ == "__main__":

    param = basicVariables()
    print(param.B_minMax," " , param.I_minMax," ", param.K," ", param.nbsimulOpt)

    # for i in range(1,len(param.discrertize_d)):
    #     print param.discrertize_d[i]



# Copyright (C) 2017 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptReg 
import StOptGeners

# unit test to show how to store some regression object and basis function coefficients associated
##################################################################################################

def  createData():

	X1=np.arange(0.0 , 2.2 , 0.01 )
	X2=np.arange(0.0 , 1.1 , 0.005 )
	Y=np.zeros((len(X1),len(X2)))
	for i in range(len(X1)):
	    for j in range(len(X2)):
	        if i < len(X1)//2:
	            if j < len(X2)//2:
	                Y[i,j]=X1[i]+X2[j]
	            else:
	               Y[i,j]=4*X1[i]+4*X2[j]
	        else:
	            if j < len(X2)//2:
	                Y[i,j]=2*X1[i]+X2[j]
	            else:
	                Y[i,j]=2*X1[i]+3*X2[j]

	XX1, XX2 = np.meshgrid(X1,X2)  
	Y=Y.T

	r,c = XX1.shape

	X =  np.reshape(XX1,(r*c,1))[:,0]
	I =  np.reshape(XX2,(r*c,1))[:,0]
	Y =  np.reshape(Y,(r*c,1))[:,0]

	xMatrix = np.zeros((2,len(X)))
	xMatrix[0,:] = X
	xMatrix[1,:] = I

	return xMatrix, Y




# main

xMatrix, y = createData()

# 2 dimensional regression 2 by 2 meshes
nbMesh = np.array([2,2],dtype=np.int32)
regressor = StOptReg.LocalLinearRegression(False,xMatrix,nbMesh)

# coefficients
coeff = regressor.getCoordBasisFunction(y)

print("Regressed coeff", coeff) 


# store them in a matrix
coeffList = np.zeros(shape=(1,3*2*2))
coeffList[0,:]=coeff.transpose()


# archive write for  regressors
archiveWriteForRegressor =StOptGeners.BinaryFileArchive("archive","w")

# store
step =1
archiveWriteForRegressor.dumpSome2DArray("RegCoeff",step,coeff)
archiveWriteForRegressor.dumpSomeRegressor("Regressor",step,regressor)
#  store a one D array
name1 = "AllSteps"
name2 = "Product"
for i in range(5):
        x = i*np.ones(10)
        archiveWriteForRegressor.dumpSome1DArray(name1, name2, x)

# archive Read for  regressors
archiveReadForRegressor =StOptGeners.BinaryFileArchive("archive","r")

# get back
values = archiveReadForRegressor.readSome2DArray("RegCoeff",step)
reg = archiveReadForRegressor.readSomeRegressor("Regressor",step)
print("Regressed coeff ", values)
print("Reg",reg)

for i in range(5):
        x =archiveReadForRegressor.readSome1DArray(name1, name2, i)
        print("X ", x)





import sys
import os
import fileinput
import re

stringToReplace  = sys.argv[1]
print " stringToReplace " , stringToReplace
newString  = sys.argv[2]
print " newString " , newString
extention =sys.argv[3]
print " extention " , extention

pathh = ['../StOpt', '../test']
for xpath in pathh:
    for path, subdirs, files   in os.walk(xpath):
        for name in files:
            if name.endswith("."+extention):
                print " file " , name
                fp =open(path+"/"+name.rstrip(),'r')   
                dlines = fp.readlines()    
                fp.close()
                fp =open(path+"/"+name.rstrip(),'w')  
                for l in dlines:  
                    l= re.sub(stringToReplace,newString,l)
                    fp.write(l)
                fp.close()


 



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import matplotlib.pyplot as plt
import microgridDEA.parameters as bv
import time

def penalty(param,inventory):
    return ((inventory - param.I0)**2)*0

def finalValue(param,regParams,demand,inventory):

    if regParams.rmctype == 'regress now 2D':
        nsim = len(demand)
        value = np.zeros((nsim,param.H+1))
        for q in range(param.H+1):
            value[:,q] = penalty(param,inventory)
        return value
    
    elif regParams.rmctype == 'gd':
        nsim = len(demand)
        gridI = len(inventory)

        value = np.zeros((nsim,gridI*(param.H+1)))
        for i in range(gridI):
            for q in range(param.H+1):
                value[:,q*gridI + i] = penalty(param,inventory[i])
        return value

def continuationVal(contValObject,x0,i0,q,nextInventory,nextq, type):

    if type=='regress now 2D':
        return contValObject.getValue([nextq],[x0,nextInventory])

    if type=='gd':
        return contValObject.getValue([nextInventory, nextq],[x0])


def currentCost(param,currentSt, control, switchCost):

    # assuming the cost function is of the type 
    # c1*d^a + c2*1{st>0} + c3*st*1{st<0}    
    cost = param.c1*(control**param.a)*param.dt + (param.c2*(currentSt if currentSt>0 else 0) + param.c3*(-currentSt if currentSt<0 else 0))*param.dt    
    if switchCost == "No":
        return cost
    else:
        return cost + param.K



def calculateCost(param,currentSt, contValObject, x0,i0,q,control,i1,q1,ohc,switchCost="No"):
    return currentCost(param,currentSt, control, switchCost) + max(continuationVal(contValObject,x0,i0,q,i1,q1,'gd'),0)




# one step optimization
def findOptimalControl(x0, i0, q, contValObject, param,regParams):

    B_max = param.B_minMax[1]
    B_min = param.B_minMax[0]

    I_max = param.I_minMax[1]
    I_min = param.I_minMax[0]

    maxOutputBattery = ((i0 - I_min)/param.dt)
    maxInputBattery = - (I_max - i0)/param.dt
    maxOutputBattery =  (B_max if maxOutputBattery>B_max else maxOutputBattery)
    maxInputBattery = (B_min if maxInputBattery<B_min else maxInputBattery)


    possibleControl = np.zeros((1,2))
    possibleControl[:,0] = 0

    possibleControl[:,1] = x0*(x0>0) + np.abs(maxInputBattery)


    demandExContol = x0 - possibleControl; # nparray of shape (1 rows and 2 columns)
    

    St = np.where((demandExContol<=maxOutputBattery) & (demandExContol>=maxInputBattery),0,demandExContol)
    St = np.where((demandExContol>maxOutputBattery),demandExContol - maxOutputBattery,St)
    St = np.where((demandExContol<maxInputBattery),demandExContol - maxInputBattery,St)

    Bt = demandExContol - St

    nextInventory = i0 - Bt*param.dt

    nextq = np.ones_like(demandExContol)
    nextq[:,0] = 0

    contVal = np.zeros_like(demandExContol)

    contVal[:,0] = continuationVal(contValObject,x0,i0,q,nextInventory[0,0],nextq[0,0], regParams.rmctype) #contValObject[0].getYhatOutSample(x0,nextInventory[:,0])
    contVal[:,1] = continuationVal(contValObject,x0,i0,q,nextInventory[0,1],nextq[0,1], regParams.rmctype) #contValObject[1].getYhatOutSample(x0,nextInventory[:,1])

    cost = param.c1*(possibleControl**param.a)*param.dt + param.c2*np.where(St>0, St, 0)*param.dt  + param.c3*np.where(St<0, -St, 0)*param.dt + contVal

    if q==0:
        switchCost = np.zeros_like(demandExContol) + param.K
        switchCost[:,0] = 0
        cost=cost+switchCost


    indx = possibleControl[:,1]<0.000001
    cost[indx,1]=10**11

    indx = np.argmin(cost,1)
    
    return cost[0, indx], possibleControl[0,indx], St[0,indx], nextInventory[0,indx], Bt[0,indx]


def optimization(contValObject, demand, inventory, param, regParamsTminus1, regParamsT):

    regime = np.arange(0,param.H+1,1)
    lc = len(regime)
    nsim = len(demand)

    if regParamsTminus1.rmctype == 'gd':
        gridI = len(inventory)
    elif regParamsTminus1.rmctype == 'regress now 2D':
        gridI = 1

    value = np.zeros((nsim,lc*gridI))
    policy_d = np.zeros((nsim,lc*gridI))
    
    StOuput = np.zeros((nsim,lc*gridI))
    StOuput[:,:]=np.nan

    policy_m = np.zeros((nsim,lc*gridI))
    policy_m[:,:]=None

    # iteration over each sample of residual demand
    for i in range(nsim):

        # iteration over each regime
        for q in regime:
            
            if regParamsTminus1.rmctype == 'gd':
                # if rmctype is grid-discretization (gd) then iterate over grid levels.
                for j in range(gridI):

                    value[i,q*gridI + j], policy_d[i,q*gridI + j], StOuput[i,q*gridI + j], _, _  = findOptimalControl(demand[i], inventory[j], q, contValObject, param, regParamsT)

            elif regParamsTminus1.rmctype == 'regress now 2D':

                value[i,q], policy_d[i,q], StOuput[i,q], _, _  = findOptimalControl(demand[i], inventory[i], q, contValObject, param, regParamsT)


    return value,policy_d, StOuput             





# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import matplotlib
import matplotlib.pyplot as plt

import microGrid.parameters as bv
import microGrid.simulateState as simX
import microGrid.valueNext as fv


import StOptGrids
import StOptReg
import StOptGeners


def currentCost(param,St, dt):
    return (param.c1*(dt**param.a))*param.dt,  param.c2*np.where(St>0,St,0)*param.dt + param.c3*np.where(St<0,-St,0)*param.dt


def forwardSimulations():

    outSampleSimul = 10000
    maturityOutsample = 10

    param = bv.basicVariables(maturityOutsample)

    condExpFilePath = param.filetoDump

    X0 = np.zeros(outSampleSimul)
    X0[:] = 0  #initial state for residual demand

    Xt, _ = simX.residualDemand(param, param.nstep, outSampleSimul, param.maturity, X0, param.rmctype, "outsample")

    r,c = Xt.shape

    It = np.zeros((r,c)) # inventory evolution
    Dt = np.zeros((r,c), dtype=np.int) # regime 
    Dt[:,0] = 0  #initial regime
    
    St = np.zeros((r,c)) # stores Xt - dt - Bt
    Bt = np.zeros((r,c)) # battery output
    dt = np.zeros((r,c)) # diesel output
    costTrue = np.zeros((r,c)) # cost


    It[:,0] = param.I0 #initial state for battery inventory

    archiveToRead =  StOptGeners.BinaryFileArchive(condExpFilePath,"r")

    for iSteps in range(param.nstep):

        contValues = archiveToRead.readGridAndRegressedValue(39,"toStore")

        for sim in range(outSampleSimul):
            _, dt[sim, iSteps], St[sim, iSteps], It[sim, iSteps+1] ,Bt[sim, iSteps] = fv.findOptimalControl(Xt[sim,iSteps], It[sim,iSteps], Dt[sim,iSteps], contValues[0], param)

        Dt[:, iSteps+1] = np.where(dt[:, iSteps] > 0, 1, 0)

        costTrue[:,iSteps] = np.where((Dt[:,iSteps]==0) & (Dt[:,iSteps+1]==1),param.K,0)


    DieselCost, ArtificalCost = currentCost(param,St, dt) 
    costTrue = costTrue + DieselCost
    costTrue[:,-1] = fv.penalty(param,It[:,-1])



    cumCostTrue = np.cumsum(costTrue,1)

    print("Cost: Total", np.mean(np.sum(costTrue+ArtificalCost,1)), "true", np.mean(np.sum(costTrue,1)), "artifical", np.mean(np.sum(ArtificalCost,1)))
    print("std: Total", np.std(np.sum(costTrue+ArtificalCost,1))/np.sqrt(outSampleSimul), "true", np.std(np.sum(costTrue,1))/np.sqrt(outSampleSimul), "artifical", np.std(np.sum(ArtificalCost,1))/np.sqrt(outSampleSimul))

    # # plots to show curtailed energy and proportion of blackouts

    # St1 = np.where(St<-10**-9,1,0)
    # plt.plot(np.sum(St1,0)[:-2]/St1.shape[0])
    # plt.show()


    # St1 = np.where(St>10**-9,1,0)
    # plt.plot(np.sum(St1,0)[:-2]/St1.shape[0])
    # plt.show()
    
if __name__ == "__main__":
    forwardSimulations()


# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import StOptGrids
import StOptGlobal
import Utils
import Simulators as sim
import Optimizers as opt

# unit test mean reverting and gas
#################################

class MeanRevertingAndGas(unittest.TestCase):

    def test_(self):
        # grids of times values
        time = StOptGrids.OneDimRegularSpaceGrid(0,1./10,10)
        # values
        values = []
        iPeriod = 52
        for i in range(11):
             values.append(2. + math.sin(math.pi * i * iPeriod) / 10.)
        # future curve
        fut  = Utils.FutureCurve(time,values)
        # mean revering simulator
        sigma = np.zeros(1) + 0.2 # volatility
        mr = np.zeros(1) + 0.05 # mean reverting
        T =1. # maturity
        nbStep = 10 # number of time steps
        nbSimul =100 # number of simulations
        bForward =1 # forward simulation
        r= 0. # interest rate
        simulator = sim.MeanRevertingSimulator(fut,sigma,mr,r,T,nbStep,nbSimul,bForward)
        # get number of simulation
        print(simulator.getNbSimul())
if __name__ == '__main__': 
    unittest.main() 



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptGrids 
import StOptReg
import StOptGlobal
import StOptGeners


def DynamicProgrammingByRegressionHighLevel(p_grid, p_optimize, p_regressor, p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump) :
    
    # from the optimizer get back the simulation
    simulator = p_optimize.getSimulator()
    # final values
    fin = StOptGlobal.FinalStepDP(p_grid, p_optimize.getNbRegime())
    valuesNext = fin.set(p_funcFinalValue, simulator.getParticles())
    ar = StOptGeners.BinaryFileArchive(p_fileToDump, "w")
    nameAr = "Continuation"
    nsteps =simulator.getNbStep()
    # iterate on time steps
    for iStep in range(nsteps) :
        asset = simulator.stepBackwardAndGetParticles()
        # conditional expectation operator
        if iStep == (simulator.getNbStep() - 1):
            p_regressor.updateSimulations(True, asset)
        else:
            p_regressor.updateSimulations(False, asset)
        
        # transition object
        transStep = StOptGlobal.TransitionStepRegressionDP(p_grid, p_grid, p_optimize)
        valuesAndControl = transStep.oneStep(valuesNext, p_regressor)
        transStep.dumpContinuationValues(ar, nameAr, nsteps - 1 -iStep, valuesNext, valuesAndControl[1], p_regressor)
        valuesNext = valuesAndControl[0]
        
    # interpolate at the initial stock point and initial regime
    return (p_grid.createInterpolator(p_pointStock).applyVec(valuesNext[p_initialRegime])).mean()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest
import random
import math


# unit test for continuation values
##################################

class testContValues(unittest.TestCase):

    # test a regular grid  for stocks and a local function basis for regression
    def testSimpleGridsAndRegressor(self):
        import StOptGrids
        import StOptReg
        # low value for the meshes
        lowValues =np.array([1.,2.,3.],dtype=float)
        # size of the meshes
        step = np.array([0.7,2.3,1.9],dtype=float)
        # number of steps
        nbStep = np.array([3,2,4], dtype=np.int32)
        # create the regular grid
        #########################
        grid = StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)
        # simulation
        nbSimul =10000
        np.random.seed(1000)
        x = np.random.uniform(-1.,1.,size=(1,nbSimul));
        # mesh
        nbMesh = np.array([16],dtype=np.int32)
        # Create the regressor
        #####################
        regressor = StOptReg.LocalLinearRegression(False,x,nbMesh)
        # regressed values
        toReal = (2+x[0,:]+(1+x[0,:])*(1+x[0,:]))
        # function to regress
        toRegress = toReal + 4*np.random.normal(0.,1,nbSimul)
        # create a matrix (number of stock points by number of simulations)
        toRegressMult = np.zeros(shape=(len(toRegress),grid.getNbPoints()))
        for i in range(toRegressMult.shape[1]):
           toRegressMult[:,i] = toRegress
        # Now create the continuation object
        ####################################
        contOb = StOptReg.ContinuationValue(grid,regressor,toRegressMult)
        # get back the regressed values at the point stock
        ptStock=  np.array([1.2,3.1,5.9],dtype=float)
        regressValues = contOb.getAllSimulations(ptStock)
        # do the same with an interpolator
        interp = grid.createInterpolator(ptStock)
        regressValuesInterp = contOb.getAllSimulations(interp)
        # test create of an interpoaltion object mixing grids for stocks and regression for uncertainties
        #################################################################################################
        gridAndRegressed = StOptReg.GridAndRegressedValue(grid,regressor,toRegressMult)
        # get back the regressed value for a point stock and an uncertainty
        valRegressed = gridAndRegressed.getValue(ptStock, x[:,0])
        # Now test simulations one by one
        #################################
        for i in range(nbSimul):
            regressed = gridAndRegressed.getValue(ptStock,x[:,i] )
            diff = regressed -regressValues[i]
            self.assertAlmostEqual(diff,0., 7, "test regression simulation by simulation")

    # test some mapping of GneralSpaceGrid
    def testGeneralGridInheritance(self):
        from StOptGrids import GeneralSpaceGrid, RegularSpaceGrid
        from StOptReg import LocalLinearRegression, ContinuationValue

        x = np.random.randn(5)
        regressor = LocalLinearRegression([1])
        
        regular = RegularSpaceGrid(np.array([0.]), np.array([0.5]), np.array([3]))
        ContinuationValue(regular, regressor, x) 
        
        general = GeneralSpaceGrid([[0., 1., 1.2, 1.5]])
        ContinuationValue(general, regressor, x) 
   
 
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import importlib
import numpy as np
import unittest
import StOptGrids
import StOptReg
import StOptGlobal
import StOptGeners
import Utils
import Simulators as sim
import Optimizers as opt

# unit test for global shape
############################

class OptimizerConstruction(unittest.TestCase):

    def test(self):
        # test MPI
        moduleMpi4Py=importlib.util.find_spec('mpi4py')
        if (moduleMpi4Py is not None):
            from mpi4py import MPI
        initialValues = np.zeros(1,dtype=float) + 1.
        sigma = np.zeros(1) + 0.2
        mu = np.zeros(1) + 0.05
        corr = np.ones((1,1),dtype=float)
        # number of step
        nStep = 30
        # exercise dates
        dates = np.linspace(0., 1., nStep + 1)
        T= dates[len(dates) - 1]
        nbSimul = 10 # simulation number (optimization and simulation)
        # simulator
        ##########
        bsSim = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, T, len(dates) - 1, nbSimul, False)
        strike = 1.
        # Pay off
        payOff= Utils.BasketCall(strike)
        # optimizer
        ##########
        N = 3   # number of exercise dates
        swiOpt = opt.OptimizerSwingBlackScholes(payOff,N)
        # link simulator to optimizer
        swiOpt.setSimulator(bsSim)
        # archive
        ########
        ar = StOptGeners.BinaryFileArchive("Archive","w")
        # regressor
        ##########
        nMesh = np.array([1])
        regressor = StOptReg.LocalLinearRegression(nMesh)
        # Grid
        ######
        # low value for the meshes
        lowValues =np.array([0.],dtype=float)
        # size of the meshes
        step = np.array([1.],dtype=float)
        # number of steps
        nbStep = np.array([N], dtype=np.int32)
        gridArrival =  StOptGrids.RegularSpaceGrid(lowValues,step,nbStep)
        gridStart   =  StOptGrids.RegularSpaceGrid(lowValues,step,nbStep-1)
        # pay off function for swing
        ############################
        payOffBasket = Utils.BasketCall(strike);
        payoff = Utils.PayOffSwing(payOffBasket,N)
        dir(payoff)
        # final step
        ############
        asset =bsSim.getParticles()
        fin = StOptGlobal.FinalStepDP(gridArrival,1)
        values = fin.set( payoff,asset)
        # transition time step
        #####################
        # on step backward and get  asset
        asset = bsSim.stepBackwardAndGetParticles()
        # update regressor
        regressor.updateSimulations(0,asset)
        transStep = StOptGlobal.TransitionStepRegressionDP(gridStart,gridArrival,swiOpt)
        valuesNextAndControl=transStep.oneStep(values,regressor)
        transStep.dumpContinuationValues(ar,"Continuation",1,valuesNextAndControl[0],valuesNextAndControl[1],regressor)
        # simulate time step
        ####################
        nbSimul= 10
        vecOfStates =[] # state of each simulation
        for i in np.arange(nbSimul):
            # one regime, all with same stock level (dimension 2), same realization of simulation (dimension 3)
            vecOfStates.append(StOptGlobal.StateWithStocks(1, np.array([0.]) , np.zeros(1)))
        arRead = StOptGeners.BinaryFileArchive("Archive","r")
        simStep = StOptGlobal.SimulateStepRegression(arRead,1,"Continuation",gridArrival,swiOpt)
        phi = np.zeros((1,nbSimul))
        VecOfStateNext = simStep.oneStep(vecOfStates, phi)

if __name__ == '__main__': 
    unittest.main() 



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
        
class basicVariables:


    def __init__(self, maturity=10):

        self.K = 10 #switching cost

        self.I0=5 #initial inventory

        self.I_minMax = [0,10] #Level of charge of the battery min max in KW
        self.B_minMax =  [-6,6] #battery min max output, positive implies output and negative implies input

        self.H = 1 # represents the number of regimes which in this case is 2, as a result H=1

        # assuming the cost function is of the type 
        # c1*d^a + c2*1{st>0} + c3*st*1{st<0}
        c =[1,10000,5] # assuming the cost function is of the type , c[0]*d^a + c[1]*1{st>0} + c[2]*1{st<0}(-st)
        self.a = 0.9
        self.c1 = c[0]
        self.c2 = c[1]
        self.c3 = c[2]


        self.maturity = maturity #unit in hours
        self.dt = 15/60 #unit in hours
        self.nstep = int(self.maturity/self.dt) #number of macro steps


        # demand parameters
        self.lambd = 0.5 #mean reversion rate
        self.sigma = 2 #volatility
        self.ampl = 6 
        self.mu = 0

        # we bound the maximum demand. This parameter defines the maximum possible demand.
        self.DemandUB = 10

        # file to dump the continuation values.
        self.filetoDump = "condExp"
        # file to dump the regression parameters.        
        self.regParamDump = 'regressionParameters.pkl'

        # parameter defines when to change the type of regression. 
        self.designChangePnt = 30


# regression related parameters
class regressionType:


    def __init__(self,rmctype='regress now 2D'):

        #number of meshes on demand dimension
        self.meshX = 4 
        #number of meshes on inventory dimension
        self.meshI = 10 
        # type of regression monte carlo i.e grid discretization or 2D regressions.
        self.rmctype = rmctype 
        
        if self.rmctype == 'gd':
            self.regType = 'piecewiseLinear' #regression type.
            # number of simulations
            self.nbsimulOpt = 500*self.meshX 

            #number of unique design sites.
            self.nbUniqueSite = 500*self.meshX 
            #batch size.     
            self.batchSize = 1  

            # exit the code if the product of the batch size and unique sites is not equal to number of simulations.
            if (self.nbUniqueSite*self.batchSize != self.nbsimulOpt):
                print("wrong budget specification: batch size and sites don't match the total simulations")
                exit(0)
            

        elif self.rmctype == 'regress now 2D':
            self.regType =  'globalPolynomial'  #'piecewiseLinear' ##'kernel' 
            self.nbsimulOpt = 10000

            #number of unique design sites.
            self.nbUniqueSite = 1000
            #batch size.
            self.batchSize = 10

            # exit the code if the product of the batch size and unique sites is not equal to number of simulations.
            if (self.nbUniqueSite*self.batchSize != self.nbsimulOpt):
                print("wrong budget specification: batch size and sites don't match the total simulations")
                exit(0)

            self.degree= 2  # used only with regType = 'globalPolynomial'

            # used only with regType = 'kernel'
            self.bandwidth = 0.2
            # factor for the number of points
            self.factPoint=1.

        



# Copyright (C) 2020 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import unittest

# unit test for fast CDF
#########################

class testFastCDFOnSample(unittest.TestCase):
    def test2D(self):
        import StOptCDF
        
        nbSimul = 10000
        x = np.random.normal(size=(2,nbSimul));
        y =  np.ones([nbSimul])

        #  CDF at sample points
        cdf= StOptCDF.fastCDFOnSample(x,y)
        
        print("CDF", cdf[:10], "And " , cdf[-10:])
if __name__ == '__main__': 
    unittest.main()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptSDDP
import StOptSDDPUnitTest
import numpy as np
import unittest
import StOptGeners

# unitest for SDDP framework
############################


class FakeOptimizer(unittest.TestCase):

    def test(self):

        # Create a fake optimizer depending of a fake simulator
        simul =StOptSDDPUnitTest.SimulTest()
        optim= StOptSDDPUnitTest.OptimizeTest(simul)

        # final cuts
        finCut =StOptSDDP.SDDPFinalCut()

        # SDDP problem parameters
        nbSimulCheckForSimu = 1
        initialState = np.array([1.,2.])
        dates  = np.array([1.,2.])
        nbMesh= np.array([1,1],np.int32)
        nameRegressor =" "
        nameCut = " "
        nameVisitedStates = " "
        iter= 1
        accuracy=0.01
        nStepConv=10
        # test backward and forward mapping
        sol = StOptSDDPUnitTest.backwardForwardSDDPTestMapping(optim,nbSimulCheckForSimu,initialState,finCut,dates,nbMesh,nameRegressor,nameCut,nameVisitedStates,iter,accuracy,nStepConv)
        


class LocalLinearRegressionForSDDP(unittest.TestCase):

    def test(self):

        # Create a local linear regressor for SDDP
        particles= np.array([[1.,2.],[3.,3.]],np.float64)
        mesh = np.array([1,2],np.int32)
                           
        regressor =StOptSDDP.LocalLinearRegressionForSDDP(1,particles,mesh)
        
        archiveToWrite = StOptGeners.BinaryFileArchive("MyArchive","w")

        # dump SDDP regressor
        archiveToWrite.dump(regressor)
        # visited state
        aVisitedState = StOptSDDP.SDDPVisitedStates(1)
        # a state
        aState =  np.array([3.,4.],np.float64)
        # a particle
        aParticle = np.array([1.,2.],np.float64)
        aVisitedState.addVisitedState(aState, aParticle,regressor)
        # dump visited state
        archiveToWrite.dump(aVisitedState)

        # test backward SDDP mapping
        simul =StOptSDDPUnitTest.SimulTest()
        optim= StOptSDDPUnitTest.OptimizeTest(simul)
        # get backward base simulator
        simBase = optim.getSimulatorBackward()
        dates = np.array([1.],np.float64)
        initialState=  np.array([1.],np.float64)
        finalCut =StOptSDDP.SDDPFinalCut()
        archiveRegresssor = StOptGeners.BinaryFileArchive("MyArchiveReg","w")
        archiveCut  = StOptGeners.BinaryFileArchive("MyArchiveCut","r")
        a = StOptSDDPUnitTest.backwardSDDPTestMapping(optim,simBase,dates,initialState,finalCut,archiveRegresssor,"MyString",archiveCut)

        # test backward mapping
        b= StOptSDDPUnitTest.forwardSDDPTestMapping(optim,simBase,dates,initialState,finalCut,1,archiveRegresssor,archiveCut,"MyString")
        
        
if __name__ == '__main__': 
    unittest.main() 



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np


class FinalStepDP:
    
    def __init__(self, p_pGridCurrent, p_nbRegime):
        
        self.m_pGridCurrent = p_pGridCurrent
        self.m_nDim = p_pGridCurrent.getDimension()
        self.m_nbRegime = p_nbRegime
        
    def operator(self, p_funcValue, p_particles):
        
        finalValues = list(range(self.m_nbRegime))
        
        if self.m_pGridCurrent.getNbPoints() > 0:
            
            for iReg in range(self.m_nbRegime):
                finalValues[iReg] = np.zeros((p_particles.shape[1], self.m_pGridCurrent.getNbPoints()))
                
            # number of thread
            nbThreads = 1
            
            # create iterator on current grid treated for processor
            iterGridPoint = self.m_pGridCurrent.getGridIteratorInc(0)
            
            # iterates on points of the grid
            for iIter in range(self.m_pGridCurrent.getNbPoints()):
                iThread = 0
                
                if iterGridPoint.isValid():
                    pointCoord = iterGridPoint.getCoordinate()
                    
                    for iReg in range(self.m_nbRegime):
                        
                        #for i in range(p_particles.shape[1]):
                        finalValues[iReg][:,iterGridPoint.getCount()] = p_funcValue.set(iReg, pointCoord, p_particles)
                            
                    iterGridPoint.nextInc(nbThreads)
                    
        return finalValues



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math
import numpy as np
import unittest
import StOptGrids
import StOptReg
import StOptGlobal
import Utils
import Simulators as sim
import Optimizers as opt
import dp.DynamicProgrammingByRegressionHighLevel as dyn
import dp.SimulateRegressionControlHighLevel as srt
import imp

accuracyClose = 0.25


class testSwingOptionTest(unittest.TestCase) :

    def test_resolutionSwing(self):

        try:
            imp.find_module('mpi4py')
            found = True
        except:
            print("Not parallel module found ")
            found = False

        if found :
            from mpi4py import MPI
            world = MPI.COMM_WORLD
            initialValues = np.zeros(1) + 1.
            sigma = np.zeros(1) + 0.2
            mu = np.zeros(1) + 0.05
            corr = np.ones((1,1))
            # number of step
            nStep = 4
            # exercise date
            dates = np.linspace(0., 1., nStep + 1)
            N = 2 # number of exercises
            T = 1.
            strike = 1.
            nbSimul = 100000
            nMesh = 8
            # payoff for swing
            payOffBasket = Utils.BasketCall(strike);
            payoff = Utils.PayOffSwing(payOffBasket,N)
            nbMesh = np.zeros(1, dtype = np.int32) + nMesh
            # Grid
            lowValues = np.zeros(1)
            step = np.zeros(1) + 1.
            nbStep = np.zeros(1, dtype = np.int32) + N
            grid =  StOptGrids.RegularSpaceGrid(lowValues, step, nbStep)
            # optimizer
            optimizer = opt.OptimizerSwingBlackScholes(payOffBasket, N)
            # initial values
            initialStock = np.zeros(1)
            initialRegime = 0
            fileToDump = "CondExpSwingOptSimHL"
            # simulator
            simulatorBack = sim.BlackScholesSimulator(initialValues, sigma, mu, corr, dates[len(dates) - 1], len(dates) - 1, nbSimul, False)
            # regressor
            regressor = StOptReg.LocalLinearRegression(nbMesh)
            # link the simulations to the optimizer
            optimizer.setSimulator(simulatorBack)
            # bermudean value
            valueOptim = dyn.DynamicProgrammingByRegressionHighLevel(grid, optimizer, regressor, payoff, initialStock, initialRegime, fileToDump)
            # simulation value
            print("Optim", valueOptim)

if __name__ == '__main__':
    unittest.main()



# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import StOptSDDP
import StOptGeners
import Simulators
import SDDPOptimizers
import imp

## @package backwardForwardSDDP
## backward and forward sweep by SDDP
## Permits to execute forward and backward sweep in python
## This is the python version of backwardForwardSDDP.cpp
## author Xavier Warin

## @brief Achieve forward and backward sweep by SDDP
## @param  p_optimizer           defines the optimiser necessary to optimize a step for one simulation solving a LP
## @param  p_nbSimulCheckForSimu defines the number of simulations to check convergence
## @param  p_initialState        initial state at the beginning of simulation
## @param  p_finalCut            object of final cuts
## @param  p_dates               vector of exercised dates, last dates correspond to the final cut object
## @param  p_meshForReg          number of mesh for regression in each direction
## @param  p_nameRegressor       name of the archive to store regressors
## @param  p_nameCut             name of the archive to store cuts
## @param  p_nameVisitedStates   name of the archive to store visited states
## @param  p_iter                maximum iteration of SDDP, on return the number of iterations achieved
## @param  p_accuracy            accuracy asked , on return estimation of accuracy achieved (expressed in %)
## @param  p_nStepConv           every p_nStepConv convergence is checked
## @return backward and forward valorization
def backwardForwardSDDP(p_optimizer, p_nbSimulCheckForSimu, p_initialState, p_finalCut,p_dates,
                        p_meshForReg, p_nameRegressor, p_nameCut, p_nameVisitedStates,
                        p_iter, p_accuracy,p_nStepConv):
    
    try:
        imp.find_module('mpi4py')
        Found = True
    except:
        print("Not parallel module found ")
        Found = False
    if Found:
        from mpi4py import MPI
        world = MPI.COMM_WORLD
    simulatorForOptim = p_optimizer.getSimulatorBackward()
    simulatorForSim = p_optimizer.getSimulatorForward()
    
    if Found:
        bTask = (world.rank==0)
    else:
        bTask = True


    if bTask:
        # archive for regressors
        archiveForRegressor =StOptGeners.BinaryFileArchive(p_nameRegressor,"w")
        # archive of first set of admissible states
        archiveForInitialState = StOptGeners.BinaryFileArchive(p_nameVisitedStates,"w")
        # list of states
        vecSetOfStates= [None] * (p_dates.size-1)
         # create regressors for each date except the last
        for  idate in range(p_dates.size-2,0,-1):
            simulatorForOptim.updateDateIndex(idate)
            # initial regressor
            particlesCurrent = simulatorForOptim.getParticles()
            # regressor
            regCurrent =StOptSDDP.LocalLinearRegressionForSDDP(False,particlesCurrent,p_meshForReg)
            # store the regressor
            archiveForRegressor.dump(regCurrent)
            setOfStates = StOptSDDP.SDDPVisitedStates(regCurrent.getNbMeshTotal())
            setOfStates.addVisitedStateForAll(p_optimizer.oneAdmissibleState(p_dates[idate + 1]),regCurrent)
            # ad admissible state
            vecSetOfStates[idate] =setOfStates
                                              
        simulatorForOptim.updateDateIndex(0)
        particlesInit = simulatorForOptim.getParticles()
        regInit= StOptSDDP.LocalLinearRegressionForSDDP(True,particlesInit,p_meshForReg)
        archiveForRegressor.dump(regInit)
        setOfStates = StOptSDDP.SDDPVisitedStates(1)
        setOfStates.addVisitedStateForAll(p_initialState,regInit)
        vecSetOfStates[0]= setOfStates

        for idate in range(0,p_dates.size-1):
            archiveForInitialState.dump(vecSetOfStates[idate])


    iterMax = p_iter;
    p_iter = 0;
    accuracy = p_accuracy;
    p_accuracy = 1e10;
    # archive to read regressor : currently all processor reads the regression
    archiveReadRegressor = StOptGeners.BinaryFileArchive(p_nameRegressor, "r");
    # only create for first task
    archiveForCuts = StOptGeners.BinaryFileArchive(p_nameCut, "w+");
    # to store  all backward values
    backwardValues =[]
    # forward value
    forwardValueForConv = 0.
    istep = 0;
    # store evolution of convergence
    backwardMinusForwardPrev = 0;
    while ((accuracy < p_accuracy) and (p_iter < iterMax)):
        # increase step
        istep += 1;
        # actualize time for simulators
        simulatorForOptim.resetTime()
        simulatorForSim.resetTime()
        
        if Found:
            world.Barrier()
        # backward sweep
        backwardValues.append(StOptSDDP.backwardSDDP(p_optimizer , simulatorForOptim, p_dates,  p_initialState,
                                                     p_finalCut, archiveReadRegressor,
                                                     p_nameVisitedStates , archiveForCuts))        
        if Found:
            world.Barrier()

        if ((istep == p_nStepConv) or (p_iter == 0)):
            istep = 0
            oldParticleNb = simulatorForSim.getNbSimul()
            simulatorForSim.updateSimulationNumberAndResetTime(p_nbSimulCheckForSimu)
            bIncreaseCut = False;
            
            forwardValueForConv =  StOptSDDP.forwardSDDP(p_optimizer , simulatorForSim, p_dates, p_initialState, p_finalCut, bIncreaseCut, archiveReadRegressor,
                                                         archiveForCuts, p_nameVisitedStates)

            p_accuracy  = abs((backwardValues[p_iter] - forwardValueForConv) / forwardValueForConv)
            simulatorForSim.updateSimulationNumberAndResetTime(oldParticleNb)

            if bTask:
                print(" ACCURACY ", p_accuracy, " Backward ", backwardValues[p_iter], " Forward ", forwardValueForConv , "p_iter "  ,p_iter , " accuracy " ,  p_accuracy )
            backwardMinusForward = backwardValues[p_iter] - forwardValueForConv;
            if (p_iter > 0):
                if (backwardMinusForward * backwardMinusForwardPrev < 0):
                    if bTask:
                        print(" Curve are crossing : increase sample and simulations to get more accurate solution, decrease step for checking convergence")
                    #exit
                    p_accuracy = 0.
            backwardMinusForwardPrev = backwardMinusForward;
        elif(p_iter>0):
            backwardMinusForward = backwardValues[p_iter] - forwardValueForConv;
            if (backwardMinusForward * backwardMinusForwardPrev < 0):
                if bTask:
                    print(" Curve are crossing : increase sample and simulations to get more accurate solution, decrease step for checking convergence")
                    #exit
                p_accuracy = 0.
        p_iter += 1;
    if Found:
        world.Barrier()
    return [backwardValues[p_iter - 1], forwardValueForConv]





# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import math as mth
import utils.NormalCumulativeDistribution as cdf

# Classical call option
class CallOption :

    # Constructor
    def __init__(self) :

        return None

    # value
    # p_S        asset value
    # p_sigma   volatility
    # p_r       interest rate
    # p_strike  strike
    # p_mat     maturity
    # return option value
    def operator(self, p_S, p_sigma, p_r, p_strike, p_mat) :

        d1 = (mth.log(p_S / p_strike) + (p_r + 0.5 * p_sigma * p_sigma) * p_mat) / (p_sigma * mth.sqrt(p_mat))
        d2 = d1 - p_sigma * mth.sqrt(p_mat)
        norm = cdf.NormalCumulativeDistribution()

        return p_S * norm.operator(d1) - p_strike * mth.exp(-p_r * p_mat) * norm.operator(d2)

# Classical put option
class PutOption :

    # Constructor
    def __init__(self) :

        return None

    # value
    # p_S        asset value
    # p_sigma   volatility
    # p_r       interest rate
    # p_strike  strike
    # p_mat     maturity
    # return option value
    def operator(self, p_S, p_sigma, p_r, p_strike, p_mat) :

        d1 = (mth.log(p_S / p_strike) + (p_r + 0.5 * p_sigma * p_sigma) * p_mat) / (p_sigma * mth.sqrt(p_mat))
        d2 = d1 - p_sigma * mth.sqrt(p_mat)

        return -p_S * cdf.operator(-d1) + p_strike * mth.exp(-p_r * p_mat) * cdf.operator(-d2)



# Copyright (C) 2019 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import StOptGrids 
import StOptTree
import StOptGlobal
import StOptGeners


def DynamicProgrammingByTreeHighLevel(p_grid, p_optimize,  p_funcFinalValue, p_pointStock, p_initialRegime, p_fileToDump) :
    
    # from the optimizer get back the simulation
    simulator = p_optimize.getSimulator()
    # final values
    fin = StOptGlobal.FinalStepDP(p_grid, p_optimize.getNbRegime())
    valuesNext = fin.set(p_funcFinalValue, simulator.getNodes())
    ar = StOptGeners.BinaryFileArchive(p_fileToDump, "w")
    nameAr = "Continuation"
    nsteps =simulator.getNbStep()
    # iterate on time steps
    for iStep in range(nsteps) :
        simulator.stepBackward()
        # get back probability
        proba = simulator.getProba()
        # and connection matrix
        connected = simulator.getConnected()
        # creta tree for conditional expectation
        tree=StOptTree.Tree(proba,connected)      
        # transition object
        transStep = StOptGlobal.TransitionStepTreeDP(p_grid, p_grid, p_optimize)
        valuesAndControl = transStep.oneStep(valuesNext, tree)
        transStep.dumpContinuationValues(ar, nameAr, iStep, valuesNext, valuesAndControl[1], tree)
        valuesNext = valuesAndControl[0]
        
    # interpolate at the initial stock point and initial regime
    return (p_grid.createInterpolator(p_pointStock).applyVec(valuesNext[p_initialRegime])).mean()



# Copyright (C) 2018 The Regents of the University of California, Michael Ludkovski and Aditya Maheshwari
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)

from __future__ import division
import numpy as np
import math
import sys
import matplotlib
import matplotlib.pyplot as plt

import microgridDEA.parameters as bv
import microgridDEA.simulateState_new as simX
import microgridDEA.valueNext_new as fv
if (sys.version_info > (3, 0)):
        import pickle
else:
        import cPickle as pickle

import StOptGrids
import StOptReg
import StOptGeners


def currentCost(param,St, dt):
    return (param.c1*(dt**param.a))*param.dt,  param.c2*np.where(St>0,St,0)*param.dt + param.c3*np.where(St<0,-St,0)*param.dt


def forwardSimulations():

    outSampleSimul = 10000
    maturityOutsample = 10

    param = bv.basicVariables(maturityOutsample)

    condExpFilePath = param.filetoDump

    with open(param.regParamDump, 'rb') as input:
        regParameters = pickle.load(input)

    X0 = np.zeros(outSampleSimul)
    X0[:] = 0  #initial state for residual demand

    Xt = simX.residualDemand(param, param.nstep, outSampleSimul, param.maturity, X0, regParameters[0].rmctype, "outsample")

    r,c = Xt.shape

    It = np.zeros((r,c)) # inventory evolution
    Dt = np.zeros((r,c), dtype=np.int) # regime 
    Dt[:,0] = 0  #initial regime
    
    St = np.zeros((r,c)) # stores Xt - dt - Bt
    Bt = np.zeros((r,c)) # battery output
    dt = np.zeros((r,c)) # diesel output
    costTrue = np.zeros((r,c)) # cost


    It[:,0] = param.I0 #initial state for battery inventory

    archiveToRead =  StOptGeners.BinaryFileArchive(condExpFilePath,"r")


    for iSteps in range(param.nstep):

        contValues = archiveToRead.readGridAndRegressedValue(param.nstep-iSteps,"toStore")

        for sim in range(outSampleSimul):
            _, dt[sim, iSteps], St[sim, iSteps], It[sim, iSteps+1] ,Bt[sim, iSteps] = fv.findOptimalControl(Xt[sim,iSteps], It[sim,iSteps], Dt[sim,iSteps], contValues[0], param, regParameters[iSteps])

        Dt[:, iSteps+1] = np.where(dt[:, iSteps] > 0, 1, 0)

        costTrue[:,iSteps] = np.where((Dt[:,iSteps]==0) & (Dt[:,iSteps+1]==1),param.K,0)


    DieselCost, ArtificalCost = currentCost(param,St, dt) 
    costTrue = costTrue + DieselCost
    costTrue[:,-1] = fv.penalty(param,It[:,-1])



    cumCostTrue = np.cumsum(costTrue,1)

    print("Cost: Total", np.mean(np.sum(costTrue+ArtificalCost,1)), "true", np.mean(np.sum(costTrue,1)), "artifical", np.mean(np.sum(ArtificalCost,1)))
    print("std: Total", np.std(np.sum(costTrue+ArtificalCost,1))/np.sqrt(outSampleSimul), "true", np.std(np.sum(costTrue,1))/np.sqrt(outSampleSimul), "artifical", np.std(np.sum(ArtificalCost,1))/np.sqrt(outSampleSimul))

    # plots to show curtailed energy and proportion of blackouts

    # St1 = np.where(St<-10**-9,1,0)
    # plt.plot(np.sum(St1,0)[:-2]/St1.shape[0])
    # plt.show()


    # St1 = np.where(St>10**-9,1,0)
    # plt.plot(np.sum(St1,0)[:-2]/St1.shape[0])
    # plt.show()
    
if __name__ == "__main__":
    forwardSimulations()



# Copyright (C) 2016 EDF
# All Rights Reserved
# This code is published under the GNU Lesser General Public License (GNU LGPL)
import numpy as np
import utils.BasketOptions as bo
import simulators.BlackScholesSimulator as bs
import utils.americanOption as amer
import StOptGrids
import StOptReg as reg
import unittest


# Generic test case
# p_nDim                   dimension of the problem
# p_nbSimul                number of simulations used
# p_level                  level of the sparse grid
# p_degree                 degree for the interpolation
# p_referenceValue         reference value
# p_accuracyEqual          accuracy for equality
def americanSparse(p_nDim, p_nbSimul, p_level, p_degree, p_referenceValue, p_accuracyEqual) :
    
    initialValues = np.zeros(p_nDim) + 1.
    sigma = np.zeros(p_nDim) + 0.2
    mu = np.zeros(p_nDim) + 0.05
    corr = np.zeros((p_nDim, p_nDim))
    T = 1.
    nDate = 10
    np.fill_diagonal(corr, 1.)
    strike = 1.

    # simulator
    simulator = bs.BlackScholesSimulator(initialValues, sigma, mu, corr, T, nDate, p_nbSimul, False)
    # payoff
    payoff = bo.BasketPut(strike)
    # regressor
    weight = np.zeros(p_nDim) + 1.
    regressor = reg.SparseRegression(p_level, weight, p_degree)
    # bermudean value
    value = amer.resolution(simulator, payoff, regressor)
    return value
        
# test cases
class testAmericanOptionForSparse(unittest.TestCase):
    
    def test_americanSparseBasket1D(self):
        
        # dimension
        nDim = 1
        nbSimul = 100000
        level = 5
        referenceValue = 0.06031
        accuracyEqual = 0.2
    
        val = americanSparse(nDim, nbSimul, level, 1, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 2, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 3, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        
    def test_americanSparseBasket2D(self):
        
        # dimension
        nDim = 2
        nbSimul = 100000
        level = 5
        referenceValue = 0.03882
        accuracyEqual = 0.5
    
        val = americanSparse(nDim, nbSimul, level, 1, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 2, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 3, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        
    def test_americanSparseBasket3D(self):
        
        # dimension
        nDim = 3
        nbSimul = 100000
        level = 5
        referenceValue = 0.02947
        accuracyEqual = 0.5
    
        val = americanSparse(nDim, nbSimul, level, 1, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 2, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 3, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        
    def test_americanSparseBasket4D(self):
        
        # dimension
        nDim = 4
        nbSimul = 100000
        level = 5
        referenceValue = 0.02404
        accuracyEqual = 1.
    
        val = americanSparse(nDim, nbSimul, level, 1, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 2, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        val = americanSparse(nDim, nbSimul, level, 3, referenceValue, accuracyEqual)
        self.assertAlmostEqual(val, referenceValue, None, None, accuracyEqual)
        
if __name__ == '__main__':
    
    unittest.main()
